from typing import Optional, Tuple

from vbeam.fastmath import backend_manager
from vbeam.fastmath import numpy as np
from vbeam.fastmath.traceable import traceable_dataclass
from vbeam.scan.advanced.base import WrappedScan
from vbeam.scan.base import CoordinateSystem
from vbeam.scan.sector_scan import SectorScan
from vbeam.scan.util import scan_convert
from vbeam.util.vmap import vmap_all_except


def _get_scan_converted_points_and_indices(
    base_scan: SectorScan,
) -> Tuple[np.ndarray, np.ndarray]:
    # Scan convert a grid of ones to get a mask of the points that are mapped to the
    # cartesian grid.
    mask = scan_convert(np.ones(base_scan.shape), base_scan)
    mask = (mask == 1).reshape((base_scan.num_points,))

    # Calculate the scan converted points.
    points = base_scan.get_points(flatten=False)
    scan_converted_points = scan_convert(points, base_scan, 0, 1)
    scan_converted_points = scan_converted_points.reshape((base_scan.num_points, 3))

    # We must use Numpy for this because masking out values is hard on GPUs.
    with backend_manager.using_backend("numpy"):
        # Mask out the points and indices
        mask = np.array(mask)
        points = np.array(scan_converted_points)[mask]
        indices = np.arange(base_scan.num_points)[mask]

    # Wrap in array to ensure that they use the active backend.
    return np.array(points), np.array(indices)


@traceable_dataclass(("_points", "_indices"), ("_base_scan",))
class ScanConvertedSectorScan(WrappedScan):
    def __init__(
        self,
        _base_scan: SectorScan,
        _points: Optional[np.ndarray] = None,
        _indices: Optional[np.ndarray] = None,
    ):
        if _points is None or _indices is None:
            _points, _indices = _get_scan_converted_points_and_indices(_base_scan)

        self._base_scan = _base_scan
        self._points = _points
        self._indices = _indices

    def get_points(self, flatten: bool = True) -> np.ndarray:
        if not flatten:
            raise ValueError(
                f"{self.__class__.__name__} only supports getting flattened points."
            )
        return self._points

    def unflatten(self, imaged_points: np.ndarray, points_axis: int = -1) -> np.ndarray:
        image = np.zeros((self.base_scan.num_points,), dtype=imaged_points.dtype)

        def unflatten_1(imaged_points_1: np.ndarray):
            return np.add.at(image, self._indices, imaged_points_1)

        unflatten_all = vmap_all_except(unflatten_1, axis=points_axis)
        return self.base_scan.unflatten(unflatten_all(imaged_points), points_axis)

    @property
    def num_points(self) -> Tuple[int, ...]:
        return self._indices.shape[-1]

    @property
    def coordinate_system(self) -> CoordinateSystem:
        return CoordinateSystem.CARTESIAN

    @WrappedScan.base_scan.setter
    def base_scan(self, new_base_scan):
        if not isinstance(new_base_scan, SectorScan):
            raise ValueError("base_scan must be a SectorScan.")
        WrappedScan.base_scan.fset(self, new_base_scan)
        self._points, self._indices = _get_scan_converted_points_and_indices(
            self.base_scan
        )

    def copy(self):
        return ScanConvertedSectorScan(self.base_scan, self._points, self._indices)



from vbeam.data_importers.pyuff_importer import (
    import_pyuff,
    parse_beamformed_data,
    parse_pyuff_scan,
)
from vbeam.data_importers.setup import SignalForPointSetup

__all__ = [
    "SignalForPointSetup",
    "import_pyuff",
    "parse_beamformed_data",
    "parse_pyuff_scan",
]



from typing import Callable

import numpy
from numpy import allclose
from vbeam.fastmath import Backend


def test_add_at(np: Backend):
    a = np.zeros((5,))

    indices = np.array([], dtype="int32")
    b = np.array([])
    assert allclose(
        np.add.at(a, indices, b), [0, 0, 0, 0, 0], 0, 0
    ), "adding at no indices returns the array as-is"

    indices = np.array([0, 0, 0])
    b = np.array([1, 1, 1])
    assert allclose(
        np.add.at(a, indices, b), [3, 0, 0, 0, 0], 0, 0
    ), "adding 1 to the same index three times increases that value by 3"

    indices = np.array([0, 2, 3, 3])
    b = np.array([1.2, 3.14, 4.7, 5.1])
    assert allclose(
        np.add.at(a, indices, b), [1.2, 0.0, 3.14, 9.8, 0.0]
    ), "adding different numbers at different indices works as expected"

    indices = np.array([-1, -2, -3])
    b = np.array([1, 1, 1])
    assert allclose(
        np.add.at(a, indices, b), [0.0, 0.0, 1.0, 1.0, 1.0]
    ), "adding at negative indexes works as expected"


def test_add_at_multidimensional(np: Backend):
    a = np.zeros((5,))

    indices = np.array([[0, 1], [0, 4], [4, 4]], dtype="int32")
    b = np.array([[1, 1], [1, 1], [1, 1]])
    assert allclose(
        np.add.at(a, indices, b), [2, 1, 0, 0, 3]
    ), "the shape of indices and b doesn't matter"


def test_vmap(np: Backend, jit_able: Callable[[Callable], Callable]):
    a = np.ones((2, 3, 4))
    b = np.ones((3, 4))

    sum_first_and_last_axis = lambda a: np.sum(a, (0, -1))
    assert jit_able(np.vmap(sum_first_and_last_axis, [0]))(a).shape == (2,)
    assert jit_able(np.vmap(sum_first_and_last_axis, [1]))(a).shape == (3,)
    assert jit_able(np.vmap(sum_first_and_last_axis, [2]))(a).shape == (4,)

    sum_ab_1 = lambda a, b: a + b
    sum_ab = np.vmap(sum_ab_1, [0, None])
    sum_ab = np.vmap(sum_ab, [1, 0])
    sum_ab = np.vmap(sum_ab, [2, 1])
    assert jit_able(sum_ab)(a, b).shape == (4, 3, 2)

    sum_ab_1 = lambda a, b: a + b
    sum_ab = np.vmap(sum_ab_1, [0, 0])
    sum_ab = np.vmap(sum_ab, [1, 1])
    sum_ab = np.vmap(sum_ab, [0, None])
    assert jit_able(sum_ab)(a, b).shape == (2, 4, 3)


def test_nan_to_num(np: Backend, jit_able: Callable[[Callable], Callable]):
    original_array = np.array(numpy.array([numpy.nan, numpy.nan, 1, numpy.nan]))
    assert allclose(jit_able(np.nan_to_num)(original_array), np.array([0, 0, 1, 0]))



from abc import ABC, abstractmethod

from vbeam.fastmath import Backend, backend_manager
from vbeam.fastmath import numpy as np
from vbeam.fastmath.traceable import traceable_dataclass


class MyABC(ABC):
    @abstractmethod
    def do_thing(self):
        ...


@traceable_dataclass(data_fields=["a"])
class Item(MyABC):
    a: np.ndarray = 1.0

    def do_thing(self):
        return 2.0


@traceable_dataclass(data_fields=["item"])
class Container(MyABC):
    item: Item

    def do_thing(self):
        return 3.0


def test_tensorflow_traceable_dataclass_with_gradient():
    import tensorflow as tf

    backend_manager.active_backend = "tensorflow"

    @tf.function(jit_compile=True)
    def func(container: Container):
        return container.item.a + container.item.do_thing() + container.do_thing()

    item = Item(a=10.5)
    container = Container(item)
    with tf.GradientTape() as tape:
        tape.watch(container.item.a)
        res = func(container)

    assert res == 15.5
    assert tape.gradient(res, container.item.a) == 1.0


def test_jax_traceable_dataclass_with_gradient():
    import jax

    backend_manager.active_backend = "jax"

    @jax.jit
    @jax.value_and_grad
    def func(container: Container):
        return container.item.a + container.item.do_thing() + container.do_thing()

    item = Item(a=10.5)
    container = Container(item)
    res, container_grad = func(container)

    assert res == 15.5
    assert container_grad.item.a == 1.0


def test_numpy_traceable_dataclass():
    backend_manager.active_backend = "numpy"

    def func(container: Container):
        return container.item.a + container.item.do_thing() + container.do_thing()

    item = Item(a=10.5)
    container = Container(item)
    res = func(container)

    assert res == 15.5


def test_dataclass_fields_float_0d_tensor(np: Backend):
    """There have been some (Tensorflow-specific) problems with creating dataclasses
    with fields of scalar types and constructing them using 0-dimensional (aka scalar)
    values.
    Tensorflow has complained that it "expected float, got <tf.Tensor: shape=(), ...>."
    """

    @traceable_dataclass(data_fields=["a"])
    class Foo:
        a: float

    some_tensor = np.array([0.2, 0.1, 0.3])
    obj = Foo(a=some_tensor.min())
    assert obj.a == 0.1



from typing import Callable


def grad_for_argname(f: Callable, arg_name: str):
    """Wrap ``f`` in :func:`jax.grad` such that the gradient is taken with respect to
    ``arg_name``.

    Normally, :func:`jax.grad` can only take the gradient with respect to positional
    arguments. This function allows you to take the gradient with respect to a keyword
    argument.

    >>> import jax
    >>> def f(x, y):
    ...     return x * y
    >>> f_grad1 = jax.grad(f, argnums=1)  # The argument with index=1 is "y"
    >>> f_grad2 = grad_for_argname(f, "y")  # Here we reference it by name instead

    ``f_grad1`` and ``f_grad2`` are equivalent in this case:

    >>> f_grad1(2., 3.) == f_grad2(2., 3.)
    True
    """
    import jax

    def wrapped(**kwargs):
        if arg_name not in kwargs:
            raise ValueError(
                f"An error occurred when calculating the gradient for {f}: \
{arg_name} not found in kwargs."
            )

        def pos_arg_f(arg):
            new_kwargs = kwargs.copy()
            new_kwargs[arg_name] = arg  # Overwrite arg_name with traced value
            return f(**new_kwargs)

        f_grad = jax.grad(pos_arg_f)
        return f_grad(kwargs[arg_name])

    return wrapped


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from typing import Tuple

from vbeam.core import Apodization, ElementGeometry, WaveData
from vbeam.fastmath import numpy as np
from vbeam.fastmath.traceable import traceable_dataclass
from vbeam.util.geometry.v2 import Line

from .window import Window


@traceable_dataclass(("window", "beam_width", "array_bounds_x"))
class MLAApodization(Apodization):
    """Perform multiple line acquisition (MLA) in cartesian space.

    The number of lines acquired per transmit is determined by beam_width. If beam_width
    is the same as the delta_x of the scan (assuming of course that the scan dimensions
    are linear) then this apodization would produce approximately the same effect as
    using scanline imaging."""

    window: Window
    beam_width: float
    array_bounds_x: Tuple[float, float]

    def __call__(
        self,
        sender: ElementGeometry,
        point_position: np.ndarray,
        receiver: ElementGeometry,
        wave_data: WaveData,
    ) -> float:
        # Geometry assumes 2D points
        point_position = point_position[np.array([0, 2])]
        source = wave_data.source[np.array([0, 2])]

        array_left = np.array([self.array_bounds_x[0], 0])
        array_right = np.array([self.array_bounds_x[1], 0])
        mid_line_angle = (
            Line.passing_through(array_left, source).angle
            + Line.passing_through(array_right, source).angle
        ) / 2
        array_line = Line.passing_through(array_left, array_right)
        _, sender_position = array_line.intersect(
            Line.with_angle(source, mid_line_angle)
        )

        # The distance from point_position to the nearest point on the line that
        # intersects sender_position and wave_data.source.
        scanline = Line.passing_through(sender_position, source)
        dist = np.abs(scanline.signed_distance(point_position))
        return self.window(dist / (self.beam_width))



from typing import Callable, Optional

from spekk import Spec

from vbeam.core import (
    ElementGeometry,
    ReflectedWavefront,
    TransmittedWavefront,
    WaveData,
)
from vbeam.fastmath import numpy as np
from vbeam.util._default_values import (
    _default_point_position,
    _default_receiver,
    _default_sender,
    _default_wave_data,
)
from vbeam.wavefront.util import (
    get_reflected_wavefront_values,
    get_transmitted_wavefront_values,
)


def plot_transmitted_wavefront(
    wavefront: TransmittedWavefront,
    sender: Optional[ElementGeometry] = None,
    point_position: Optional[np.ndarray] = None,
    wave_data: Optional[WaveData] = None,
    spec: Optional[Spec] = None,
    postprocess: Optional[Callable[[np.ndarray], np.ndarray]] = None,
    ax=None,  # : Optional[matplotlib.pyplot.Axes]
):
    """Plot the (transmit) wavefront distance values using ``matplotlib``.

    Try to set reasonable default arguments if they are not given. Use ``postprocess``
    to process the values further before plotting (for example scan conversion)."""
    # Try to set helpful default values
    spec = spec if spec is not None else Spec({})
    if sender is None:
        sender, spec = _default_sender(spec)
    if point_position is None:
        point_position, spec = _default_point_position(spec)
    if wave_data is None:
        wave_data, spec = _default_wave_data(spec)

    # Calculate wavefront values
    vals = get_transmitted_wavefront_values(
        wavefront,
        sender,
        point_position,
        wave_data,
        spec,
        # We want to keep the dimensions of point_position
        spec["point_position"].tree,
    )
    if postprocess is not None:
        vals = postprocess(vals)

    # Plot wavefront values
    if ax is None:
        import matplotlib.pyplot as plt

        ax = plt
    return ax.imshow(
        vals.T,
        aspect="auto",
        extent=[
            point_position[..., 0].min(),
            point_position[..., 0].max(),
            point_position[..., 2].max(),
            point_position[..., 2].min(),
        ],
    )


def plot_reflected_wavefront(
    wavefront: ReflectedWavefront,
    point_position: Optional[np.ndarray] = None,
    receiver: Optional[ElementGeometry] = None,
    spec: Optional[Spec] = None,
    postprocess: Optional[Callable[[np.ndarray], np.ndarray]] = None,
    ax=None,  # : Optional[matplotlib.pyplot.Axes]
):
    """Plot the (receive) wavefront distance values using ``matplotlib``.

    Try to set reasonable default arguments if they are not given. Use ``postprocess``
    to process the values further before plotting (for example scan conversion)."""
    # Try to set helpful default values
    spec = spec if spec is not None else Spec({})
    if point_position is None:
        point_position, spec = _default_point_position(spec)
    if receiver is None:
        receiver, spec = _default_receiver(spec)

    # Calculate wavefront values
    vals = get_reflected_wavefront_values(
        wavefront,
        point_position,
        receiver,
        spec,
        # We want to keep the dimensions of point_position
        spec["point_position"].tree,
    )
    if postprocess is not None:
        vals = postprocess(vals)

    # Plot wavefront values
    if ax is None:
        import matplotlib.pyplot as plt

        ax = plt
    return ax.imshow(
        vals.T,
        aspect="auto",
        extent=[
            point_position[..., 0].min(),
            point_position[..., 0].max(),
            point_position[..., 2].max(),
            point_position[..., 2].min(),
        ],
    )



import operator
from abc import ABC, abstractmethod
from enum import Enum
from functools import reduce
from typing import Callable, Literal, Optional, Tuple, Union

from vbeam.fastmath import numpy as np
from vbeam.util import ensure_positive_index


class CoordinateSystem(Enum):
    CARTESIAN = "cartesian"
    POLAR = "polar"


class Scan(ABC):
    @abstractmethod
    def get_points(self, flatten: bool = True) -> np.ndarray:
        """Return the points defined by the scan, flattened to a (N, 3) array by
        default, where N is the number of points."""

    @abstractmethod
    def replace(self, *_axes: Union[np.ndarray, Literal["unchanged"]]) -> "Scan":
        "Return a copy of the scan with values replaced."

    @abstractmethod
    def update(self, *_axes: Optional[Callable[[np.ndarray], np.ndarray]]) -> "Scan":
        "Return a copy of the scan with updates applied to the given axes."

    @abstractmethod
    def resize(self, *_axes: Optional[int]) -> "Scan":
        "Return a copy of the scan where the given axes are resized."

    @property
    @abstractmethod
    def axes(self) -> Tuple[np.ndarray, ...]:
        """Return the axes of the scan.

        E.g.: if the scan is a sector scan, return the azimuth and depths axes."""

    @property
    def shape(self) -> Tuple[int, ...]:
        "Return the shape of the grid of points defined by the scan."
        return tuple([len(axis) for axis in self.axes])

    @property
    def bounds(self) -> np.ndarray:
        "Return the bounds of the axes of the scan."
        bounds = []
        for ax in self.axes:
            bounds += [ax[0], ax[-1]]
        return tuple(bounds)

    @property
    @abstractmethod
    def cartesian_bounds(self) -> np.ndarray:
        """Return the bounds in cartesian coordinates of the axes of the scan (useful
        for sector scans)."""

    @property
    @abstractmethod
    def coordinate_system(self) -> CoordinateSystem:
        """Return the coordinate system of the scan. E.g.: sector scans are in polar
        coordinates and linear scans are in cartesian coordinates."""

    def unflatten(self, imaged_points: np.ndarray, points_axis: int = -1) -> np.ndarray:
        "Unflatten a flattened array of values into the original shape of the scan."
        points_axis = ensure_positive_index(imaged_points.ndim, points_axis)
        return imaged_points.reshape(
            (
                imaged_points.shape[:points_axis]
                + self.shape
                + imaged_points.shape[points_axis + 1 :]
            )
        )

    @property
    def num_points(self) -> int:
        "The number of points in the scan."
        return reduce(operator.mul, self.shape)

    @property
    def ndim(self) -> int:
        "The number of dimensions of the scan."
        return len(self.shape)

    @property
    def is_2d(self) -> bool:
        "True if only two axes are defined."
        return self.ndim == 2

    @property
    def is_3d(self) -> bool:
        "True if all three axes are defined."
        return self.ndim == 3



from typing import Optional, Tuple

from vbeam.apodization.window import Window
from vbeam.core import Apodization, ElementGeometry, WaveData
from vbeam.fastmath import numpy as np
from vbeam.fastmath.traceable import traceable_dataclass
from vbeam.util import ensure_2d_point
from vbeam.util.geometry.v2 import Line, distance


def rtb_apodization(
    point: np.ndarray,
    array_left: np.ndarray,
    array_right: np.ndarray,
    focus_point: np.ndarray,
    minimum_aperture: float,
    maximum_aperture: Optional[float] = None,
    window: Optional[Window] = None,
):
    """Focused apodization used for retrospective transmit beamforming (RTB).

    A focused apodization has three main parts:
    - An hourglass shape constructed by the area between two lines drawn from either
      side of the array and passing through the focus point (line_left and line_right).
    - A line with a thickness given by minimum_aperture that passes through the focus
      point, to account for the fact that the hourglass apodization is infinitely
      narrow at the focus point.
    - A line with a thickness given by maximum_aperture that passes through the focus
      point that sets the maximum width of the hourglass apodization.
    """
    # Ensure that all points are 2D (only 2D is supported)
    point = ensure_2d_point(point)
    array_left = ensure_2d_point(array_left)
    array_right = ensure_2d_point(array_right)
    focus_point = ensure_2d_point(focus_point)

    # Set up geometry
    line_left = Line.passing_through(array_left, focus_point)
    line_right = Line.passing_through(array_right, focus_point)
    line_mid = Line.with_angle(focus_point, (line_left.angle + line_right.angle) / 2)
    line_mid_perpendicular = Line(point, line_mid.normal)

    # Short circuit if no window is given (same as giving a Rectangular window)
    if window is None:
        is_within_hourglass = (
            line_left.signed_distance(point) * line_right.signed_distance(point) <= 0
        )
        is_within_min_aperture = (
            np.abs(line_mid.signed_distance(point)) < minimum_aperture
        )
        is_within_max_aperture = (
            maximum_aperture is None
            or np.abs(line_mid.signed_distance(point)) < maximum_aperture
        )
        return np.logical_and(
            np.logical_or(is_within_hourglass, is_within_min_aperture),
            is_within_max_aperture,
        )

    # Calculate distances
    _, intersection_left = line_left.intersect(line_mid_perpendicular)
    _, intersection_right = line_right.intersect(line_mid_perpendicular)
    distance_left = distance(intersection_left, point)
    total_distance = distance(intersection_left, intersection_right)
    distance_mid = np.abs(line_mid.signed_distance(point))

    # Calculate apodizations
    valid = line_left.signed_distance(point) * line_right.signed_distance(point) <= 0
    hourglass_apodization = window(np.abs(distance_left / total_distance - 0.5)) * valid
    minimum_aperture_apodization = window(distance_mid / minimum_aperture)

    # Combine hourglass, minimum aperture, and maximum aperture (assumes window is
    # monotonically decreasing from 0 to 0.5).
    value = np.maximum(hourglass_apodization, minimum_aperture_apodization)
    if maximum_aperture is not None:
        value = np.minimum(value, window(distance_mid / maximum_aperture))
    return value


@traceable_dataclass(
    data_fields=("array_bounds", "array_width", "minimum_aperture", "window"),
    aux_fields=("use_parent",),
)
class RTBApodization(Apodization):
    array_bounds: Optional[Tuple[np.ndarray, np.ndarray]] = None
    array_width: Optional[float] = None
    minimum_aperture: float = 0.001  # TODO: Calculate this based on F# and wavelength
    maximum_aperture: Optional[float] = None
    window: Optional[Window] = None
    use_parent: bool = False

    def __post_init__(self):
        if self.array_bounds is None and self.array_width is None:
            raise ValueError("Either array bounds or array width must be set")

    def __call__(
        self,
        sender: ElementGeometry,
        point_position: np.ndarray,
        receiver: ElementGeometry,
        wave_data: WaveData,
    ) -> float:
        array_bounds = (
            self.array_bounds
            if self.array_bounds is not None
            else get_bounds(
                array_width=self.array_width, sender=sender, use_parent=self.use_parent
            )
        )
        return rtb_apodization(
            point_position,
            array_bounds[0],
            array_bounds[1],
            wave_data.source,
            self.minimum_aperture,
            self.maximum_aperture,
            self.window,
        )


@traceable_dataclass((("array_width", "minimum_aperture", "window", "use_parent")))
class SteppingApertureRTBApodization(Apodization):
    array_width: float
    minimum_aperture: float = 0.001  # TODO: Calculate this based on F# and wavelength
    window: Optional[Window] = None
    use_parent: bool = False

    def __call__(
        self,
        sender: ElementGeometry,
        point_position: np.ndarray,
        receiver: ElementGeometry,
        wave_data: WaveData,
    ) -> float:
        sender_element_position = np.where(
            self.use_parent, sender.parent_element.position, sender.position
        )
        sender_element_theta = np.where(
            self.use_parent, sender.parent_element.theta, sender.theta
        )
        sender_normal = np.array(
            [np.sin(sender_element_theta), 0.0, np.cos(sender_element_theta)]
        )
        array_left = (
            sender_element_position
            + np.cross(np.array([0, -1, 0]), sender_normal) * self.array_width / 2
        )
        array_right = (
            sender_element_position
            + np.cross(np.array([0, 1, 0]), sender_normal) * self.array_width / 2
        )
        return rtb_apodization(
            point_position,
            array_left,
            array_right,
            wave_data.source,
            self.minimum_aperture,
            self.window,
        )


def get_bounds(
    array_width,
    sender,
    use_parent,
):
    sender_element_position = np.where(
        use_parent, sender.parent_element.position, sender.position
    )
    sender_element_theta = np.where(
        use_parent, sender.parent_element.theta, sender.theta
    )
    sender_normal = np.array(
        [np.sin(sender_element_theta), 0.0, np.cos(sender_element_theta)]
    )
    array_left = (
        sender_element_position
        + np.cross(np.array([0, -1, 0]), sender_normal) * array_width / 2
    )
    array_right = (
        sender_element_position
        + np.cross(np.array([0, 1, 0]), sender_normal) * array_width / 2
    )
    return [array_left, array_right]



from typing import Callable

from numpy import allclose
from vbeam.core import WaveData
from vbeam.fastmath import Backend


def test_vmapping_wave_data(np: Backend, jit_able: Callable[[Callable], Callable]):
    wave_data = WaveData(source=np.array([[0, 0, 1], [0, 0, 2]]))
    inc_z = lambda wave_data: wave_data.with_updates_to(
        source=lambda source: source + np.array([0, 0, 1])
    )
    result = jit_able(np.vmap(inc_z, [0]))(wave_data)
    assert isinstance(result, WaveData)
    assert allclose(result.source, np.array([[0, 0, 2], [0, 0, 3]]))

    # dup_inc_z returns a tuple of two WaveData objects and one number
    dup_inc_z = lambda wave_data: (inc_z(wave_data), inc_z(wave_data), 1337.0)
    result = jit_able(np.vmap(dup_inc_z, [0]))(wave_data)
    assert isinstance(result[0], WaveData)
    assert isinstance(result[1], WaveData)
    assert np.is_ndarray(result[2])
    assert allclose(result[0].source, np.array([[0, 0, 2], [0, 0, 3]]))
    assert allclose(result[1].source, np.array([[0, 0, 2], [0, 0, 3]]))
    assert allclose(result[2], np.array([1337.0, 1337.0]))



from typing import Tuple, Union

from vbeam.fastmath import numpy as np
from vbeam.interpolation import FastInterpLinspace


def coherence_factor(beamformed_data: np.ndarray, receivers_axis: int):
    coherent_sum = np.abs(np.sum(beamformed_data, receivers_axis)) ** 2
    incoherent_sum = np.sum(np.abs(beamformed_data) ** 2, receivers_axis)
    num_receivers = beamformed_data.shape[receivers_axis]
    return np.nan_to_num(coherent_sum / incoherent_sum / num_receivers)


def normalized_decibels(data: np.ndarray):
    "Convert the data into decibels normalized for dynamic range."
    data_db = 20 * np.nan_to_num(np.log10(np.abs(data)))
    return data_db - data_db.max()


def _upsampling_indices(n: int, data_size: int) -> np.ndarray:
    """Return the sampling indices of data with size data_size after upsampling by n.

    The formula is a bit complicated because it is as general as possible. Check out
    the notebook `docs/tutorials/postprocess/upsampling.ipynb` for some visualizations.
    """
    is_even = n % 2 == 0
    is_odd = not is_even
    d = (n // 2 + is_odd) * 2
    return np.arange(n * data_size - d + is_odd) / n + (is_even + d - n) / (2 * n)


def upsample_by_interpolation(
    data: np.ndarray,
    n: int,
    axis: Union[int, Tuple[int, ...]] = 0,
) -> np.ndarray:
    """Upsample the data along the given axes. Multiple axes may be given."""
    # No axis has been given: upsample all axes
    if axis is None:
        axis = tuple(range(data.ndim))

    # Multiple axes have been given: upsample each axis separately
    if isinstance(axis, tuple):
        if isinstance(n, int):
            n = (n,) * len(axis)
        # Upsample each axis separately
        for axis, _n in zip(axis, n):
            data = upsample_by_interpolation(data, _n, axis)
        return data  # ...and return the data that has been upsampled along all axes

    # Only one axis has been given: upsample that axis
    sample_indices = _upsampling_indices(n, data.shape[axis])
    data = np.swapaxes(data, axis, 0)
    interpolator = FastInterpLinspace(0, 1, data.shape[0])
    interpolated_data = interpolator.interp1d(
        sample_indices, data, left=data[0], right=data[-1]
    )
    return np.swapaxes(interpolated_data, 0, axis)



from typing import Callable

from numpy import allclose
from vbeam.fastmath import Backend
from vbeam.interpolation import FastInterpLinspace


def test_interp1d(np: Backend, jit_able: Callable[[Callable], Callable]):
    interp = FastInterpLinspace(0, 1, 2)  # coords = [0, 1]
    jitted_interp1d = jit_able(
        lambda x: interp.interp1d(x, np.array([np.pi, np.pi * 2]), 42, 1337)
    )
    eval_x = np.linspace(-1, 2, 31)
    result = jitted_interp1d(eval_x)
    assert allclose(result[:10], 42), "Left padding is added"
    assert allclose(
        result[10:21], np.linspace(np.pi, np.pi * 2, 11), 1e-5, 1e-6
    ), "Coordinates within bounds are interpolated"
    assert allclose(result[21:], 1337), "Right padding is added"



from typing import Sequence, Set, Union

import vbeam.beamformers.building_blocks as building_blocks
from vbeam.beamformers.building_blocks import (
    sum_over_dimensions,
    unflatten_points,
    vectorize_over_datacube,
)
from vbeam.core import signal_for_point
from vbeam.data_importers import SignalForPointSetup
from vbeam.postprocess import coherence_factor, normalized_decibels
from vbeam.scan.advanced import ExtraDimsScanMixin
from vbeam.util.transformations import *


def get_das_beamformer(
    setup: SignalForPointSetup,
    *,
    keep_dimensions: Union[Sequence[str], Set[str]] = (),
    compensate_for_apodization_overlap: bool = True,
    log_compress: bool = True,
    scan_convert: bool = True,
):
    reduce_over_transmits = setup.spec.has_dimension("transmits") and not (
        isinstance(setup.scan, ExtraDimsScanMixin)
        and "transmits" in setup.scan.required_dimensions_for_unflatten
    )
    if reduce_over_transmits:
        keep_dimensions = set(keep_dimensions) | {"transmits"}
    return compose(
        signal_for_point,
        vectorize_over_datacube(
            setup, ignore=["transmits"] if reduce_over_transmits else []
        ),
        sum_over_dimensions(setup, keep=keep_dimensions),
        Reduce.Sum("transmits") if reduce_over_transmits else do_nothing,
        unflatten_points(setup),
        (
            building_blocks.compensate_for_apodization_overlap(setup)
            if compensate_for_apodization_overlap
            else do_nothing
        ),
        Apply(normalized_decibels) if log_compress else do_nothing,
        building_blocks.scan_convert(setup) if scan_convert else do_nothing,
    ).build(setup.spec)


def get_coherence_beamformer(
    setup: SignalForPointSetup,
    *,
    scan_convert: bool = True,
    keep_dimensions: Union[Sequence[str], Set[str]] = (),
):
    keep_dimensions = set(keep_dimensions) | {"receivers", "transmits"}
    return compose(
        signal_for_point,
        vectorize_over_datacube(setup, ignore=["transmits"]),
        sum_over_dimensions(setup, keep=keep_dimensions),
        Reduce.Sum("transmits"),
        unflatten_points(setup),
        Apply(coherence_factor, Axis("receivers")),
        building_blocks.scan_convert(setup) if scan_convert else do_nothing,
    ).build(setup.spec)



import math
from typing import Optional, Sequence

from spekk import Spec

from vbeam.core import (
    ElementGeometry,
    ReflectedWavefront,
    TransmittedWavefront,
    WaveData,
)
from vbeam.fastmath import numpy as np
from vbeam.util.transformations import *


def get_transmitted_wavefront_values(
    wavefront: TransmittedWavefront,
    sender: ElementGeometry,
    point_position: np.ndarray,
    wave_data: WaveData,
    spec: Spec,
    dimensions: Optional[Sequence[str]] = None,
    jit: bool = True,
):
    """
    Calculate and return the (transmit) wavefront distance values based on the provided
    arguments (``sender``, ``point_position``, and ``wave_data``).

    The ``dimensions`` argument determines what dimensions to keep; all others are
    summed over (except when ``dimesions`` is None where we keep all dimensions
    instead). ``spec`` describes the dimensions of the data. E.g., if ``dimensions`` is
    ``["transmits", "x", "z"]``, the result will be a 3D array with shape (Nt, Nx, Nz),
    where Nt, Nx, Nz are the sizes of the given dimensions in the data.

    Args:
        wavefront (TransmittedWavefront): The wavefront function to use.
        sender (ElementGeometry): The sender argument to ``wavefront``.
        point_position (np.ndarray): The point_position argument to ``wavefront``.
        wave_data (WaveData): The wave data argument to ``wavefront``.
        spec (Spec): A spec describing the dimensions/shape of the arguments.
        dimensions (Optional[Sequence[str]]): The dimensions to keep in the returned
            result. If it is an empty list, all dimensions are summed over. If it is
            None, all dimensions from the spec are kept.
        jit (bool): If True, the process is JIT-compiled (if the backend supports it).

    Returns:
        np.ndarray: The calculated wavefront distance values with shape corresponding
        to the dimensions defined in ``dimensions``.
    """
    kwargs = {
        "transmitted_wavefront": wavefront,
        "sender": sender,
        "point_position": point_position,
        "wave_data": wave_data,
    }

    # Return the full datacube if dimensions are not given
    # Careful! This may allocate a lot of memory.
    if dimensions is None:
        dimensions = list(spec.dimensions)

    # Define what dimensions to vmap and sum over and how
    vmap_dimensions = (
        spec["transmitted_wavefront"].dimensions
        | spec["sender"].dimensions
        | spec["point_position"].dimensions
        | spec["wave_data"].dimensions
    )
    sum_dimensions = vmap_dimensions - set(dimensions)
    reduce_sum_dimension = spec["wave_data"].dimensions & sum_dimensions
    if reduce_sum_dimension:
        # Make it only one of the dimensions (doesn't matter which one)
        reduce_sum_dimension = {reduce_sum_dimension.pop()}
        vmap_dimensions -= reduce_sum_dimension

    # Define how to calculate the wavefront values
    calculate_wavefront_values = compose(
        lambda transmitted_wavefront, *args, **kwargs: transmitted_wavefront(
            *args, **kwargs
        ),
        *[ForAll(dim) for dim in vmap_dimensions],
        Apply(np.sum, [Axis(dim) for dim in sum_dimensions - reduce_sum_dimension]),
        # [*reduce_sum_dimension][0] gets the "first element" of the set
        Reduce.Sum([*reduce_sum_dimension][0]) if reduce_sum_dimension else do_nothing,
        # Put the dimensions in the order defined by keep
        Apply(np.transpose, [Axis(dim, keep=True) for dim in dimensions]),
        # Make it run faster if `jit` is True and if the backend supports it.
        Wrap(np.jit) if jit else do_nothing,
    ).build(spec)

    # Calculate the wavefront values
    values = calculate_wavefront_values(**kwargs)
    # Return the average delay values
    values /= max(1, math.prod(spec.size(kwargs, dim) for dim in sum_dimensions))
    return values


def get_reflected_wavefront_values(
    wavefront: ReflectedWavefront,
    point_position: np.ndarray,
    receiver: ElementGeometry,
    spec: Spec,
    dimensions: Optional[Sequence[str]] = None,
    jit: bool = True,
):
    """
    Calculate and return the (receive) wavefront distance values based on the provided
    arguments (``point_position`` and ``receiver``).

    The ``dimensions`` argument determines what dimensions to keep; all others are
    summed over (except when ``dimesions`` is None where we keep all dimensions
    instead). ``spec`` describes the dimensions of the data. E.g., if ``dimensions`` is
    ``["receivers", "x", "z"]``, the result will be a 3D array with shape (Nr, Nx, Nz),
    where Nr, Nx, Nz are the sizes of the given dimensions in the data.

    Args:
        wavefront (ReflectedWavefront): The wavefront function to use.
        point_position (np.ndarray): The point_position argument to ``wavefront``.
        receiver (ElementGeometry): The receiver argument to ``wavefront``.
        spec (Spec): A spec describing the dimensions/shape of the arguments.
        dimensions (Optional[Sequence[str]]): The dimensions to keep in the returned
            result. If it is an empty list, all dimensions are summed over. If it is
            None, all dimensions from the spec are kept.
        jit (bool): If True, the process is JIT-compiled (if the backend supports it).

    Returns:
        np.ndarray: The calculated wavefront distance values with shape corresponding
        to the dimensions defined in ``dimensions``.
    """
    kwargs = {
        "reflected_wavefront": wavefront,
        "point_position": point_position,
        "receiver": receiver,
    }

    # Return the full datacube if dimensions are not given
    # Careful! This may allocate a lot of memory.
    if dimensions is None:
        dimensions = list(spec.dimensions)

    # Define what dimensions to vmap and sum over and how
    vmap_dimensions = (
        spec["reflected_wavefront"].dimensions
        | spec["point_position"].dimensions
        | spec["receiver"].dimensions
    )
    sum_dimensions = vmap_dimensions - set(dimensions)
    reduce_sum_dimension = spec["wave_data"].dimensions & sum_dimensions
    if reduce_sum_dimension:
        # Make it only one of the dimensions (doesn't matter which one)
        reduce_sum_dimension = {reduce_sum_dimension.pop()}
        vmap_dimensions -= reduce_sum_dimension

    # Define how to calculate the wavefront values
    calculate_wavefront_values = compose(
        lambda reflected_wavefront, *args, **kwargs: reflected_wavefront(
            *args, **kwargs
        ),
        *[ForAll(dim) for dim in vmap_dimensions],
        Apply(np.sum, [Axis(dim) for dim in sum_dimensions - reduce_sum_dimension]),
        # [*reduce_sum_dimension][0] gets the "first element" of the set
        Reduce.Sum([*reduce_sum_dimension][0]) if reduce_sum_dimension else do_nothing,
        # Put the dimensions in the order defined by keep
        Apply(np.transpose, [Axis(dim, keep=True) for dim in dimensions]),
        # Make it run faster if `jit` is True and if the backend supports it.
        Wrap(np.jit) if jit else do_nothing,
    ).build(spec)

    # Calculate the wavefront values
    values = calculate_wavefront_values(**kwargs)
    # Return the average delay values
    values /= max(1, math.prod(spec.size(kwargs, dim) for dim in sum_dimensions))
    return values



from typing import Callable, Literal, Optional, Tuple, Union, overload

from vbeam.fastmath import numpy as np
from vbeam.fastmath.traceable import traceable_dataclass
from vbeam.scan.base import CoordinateSystem, Scan
from vbeam.scan.util import parse_axes, polar_bounds_to_cartesian_bounds, scan_convert
from vbeam.util import _deprecations
from vbeam.util.arrays import grid
from vbeam.util.coordinate_systems import as_cartesian


@traceable_dataclass(("azimuths", "elevations", "depths", "apex"))
class SectorScan(Scan):
    azimuths: np.ndarray
    elevations: Optional[np.ndarray]  # May be None for 2D scans
    depths: np.ndarray
    apex: np.ndarray

    def get_points(self, flatten: bool = True) -> np.ndarray:
        polar_axis = self.elevations if self.is_3d else np.array([0.0])
        points = grid(self.azimuths, polar_axis, self.depths, shape=(*self.shape, 3))
        points = as_cartesian(points)
        # Ensure that points and apex are broadcastable
        apex = (
            np.expand_dims(self.apex, axis=tuple(range(1, self.ndim)))
            if self.apex.ndim > 1
            else self.apex
        )
        points = points + apex
        if flatten:
            points = points.reshape((self.num_points, 3))
        return points

    def replace(
        self,
        # "unchanged" means that the axis will not be changed.
        azimuths: Union[np.ndarray, None, Literal["unchanged"]] = "unchanged",
        elevations: Union[np.ndarray, None, Literal["unchanged"]] = "unchanged",
        depths: Union[np.ndarray, None, Literal["unchanged"]] = "unchanged",
        apex: Union[np.ndarray, None, Literal["unchanged"]] = "unchanged",
    ) -> "SectorScan":
        return SectorScan(
            azimuths=azimuths if azimuths != "unchanged" else self.azimuths,
            elevations=elevations if elevations != "unchanged" else self.elevations,
            depths=depths if depths != "unchanged" else self.depths,
            apex=apex if apex != "unchanged" else self.apex,
        )

    def update(
        self,
        azimuths: Optional[Callable[[np.ndarray], np.ndarray]] = None,
        elevations: Optional[Callable[[np.ndarray], np.ndarray]] = None,
        depths: Optional[Callable[[np.ndarray], np.ndarray]] = None,
        apex: Optional[Callable[[np.ndarray], np.ndarray]] = None,
    ) -> "SectorScan":
        return self.replace(
            azimuths(self.azimuths) if azimuths is not None else "unchanged",
            elevations(self.elevations) if elevations is not None else "unchanged",
            depths(self.depths) if depths is not None else "unchanged",
            apex(self.apex) if apex is not None else "unchanged",
        )

    def resize(
        self,
        azimuths: Optional[int] = None,
        elevations: Optional[int] = None,
        depths: Optional[int] = None,
    ) -> "SectorScan":
        if elevations is not None and self.elevations is None:
            raise ValueError(
                "Cannot resize elevations because it is not defined on this scan"
            )
        return self.replace(
            azimuths=(
                np.linspace(self.azimuths[0], self.azimuths[-1], azimuths)
                if azimuths is not None
                else "unchanged"
            ),
            elevations=(
                np.linspace(self.elevations[0], self.elevations[-1], elevations)
                if elevations is not None
                else "unchanged"
            ),
            depths=(
                np.linspace(self.depths[0], self.depths[-1], depths)
                if depths is not None
                else "unchanged"
            ),
        )

    @property
    def axes(self) -> Tuple[np.ndarray, ...]:
        if self.elevations is not None:
            return self.azimuths, self.elevations, self.depths
        else:
            return self.azimuths, self.depths

    @property
    def cartesian_bounds(self):
        """Get the bounds of the scan in cartesian coordinates. It is the same as
        bounding box of the scan-converted image."""
        if self.is_3d:
            raise NotImplementedError(
                "Cartesian bounds are not implemented for 3D scans yet.",
                "Please create an issue on Github if this is something you need.",
            )
        return polar_bounds_to_cartesian_bounds(self.bounds)

    @_deprecations.renamed_kwargs("1.0.5", imaged_points="image")
    def scan_convert(
        self,
        image: np.ndarray,
        azimuth_axis: int = -2,
        depth_axis: int = -1,
        *,  # Remaining args must be passed by name (to avoid confusion)
        shape: Optional[Tuple[int, int]] = None,
        default_value: Optional[np.ndarray] = 0.0,
        edge_handling:str ="Value",
    ):
        return scan_convert(
            image, self, azimuth_axis, depth_axis, shape=shape, default_value=default_value,edge_handling=edge_handling,
        )

    @property
    def coordinate_system(self) -> CoordinateSystem:
        return CoordinateSystem.POLAR

    def __repr__(self):
        return f"SectorScan(<shape={self.shape}>, apex={self.apex})"


@overload
def sector_scan(
    azimuths: np.ndarray,
    depths: np.ndarray,
    apex: Union[np.ndarray, float] = 0.0,
) -> SectorScan: ...  # 2D scan


@overload
def sector_scan(
    azimuths: np.ndarray,
    elevations: np.ndarray,
    depths: np.ndarray,
    apex: Union[np.ndarray, float] = 0.0,
) -> SectorScan: ...  # 3D scan


def sector_scan(*axes: np.ndarray, apex: Union[np.ndarray, float] = 0.0) -> SectorScan:
    "Construct a sector scan. See SectorScan documentation for more details."
    azimuths, elevations, depths = parse_axes(axes)
    return SectorScan(azimuths, elevations, depths, np.array(apex))



from dataclasses import dataclass

import numpy as np

from vbeam.fastmath.backend import Backend, i_at
from vbeam.fastmath.traceable import (
    get_traceable_aux_fields,
    get_traceable_data_fields,
    is_traceable_dataclass,
)


class NumpyBackend(Backend):
    @property
    def ndarray(self):
        return np.ndarray

    def zeros(self, shape, dtype=None):
        return np.zeros(shape, dtype)

    def ones(self, shape, dtype=None):
        return np.ones(shape, dtype)

    @property
    def pi(self):
        return np.pi

    def abs(self, x):
        return np.abs(x)

    def exp(self, x):
        return np.exp(x)

    def log(self, x):
        return np.log(x)

    def log10(self, x):
        return np.log10(x)
    
    def conj(self, x):
        return np.conj(x)
    
    def angle(self, x):
        return np.angle(x)

    def sin(self, x):
        return np.sin(x)

    def cos(self, x):
        return np.cos(x)

    def tan(self, x):
        return np.tan(x)

    def arcsin(self, x):
        return np.arcsin(x)

    def arccos(self, x):
        return np.arccos(x)

    def arctan2(self, y, x):
        return np.arctan2(y, x)

    def sqrt(self, x):
        return np.sqrt(x)

    def sign(self, x):
        return np.sign(x)

    def nan_to_num(self, x, nan=0.0):
        return np.nan_to_num(x, nan=nan)

    def min(self, a, axis=None):
        return np.min(a, axis=axis)
    
    def minimum(self, a, b):
        return np.minimum(a, b)
    
    def max(self, a, axis=None):
        return np.max(a, axis=axis)
    
    def maximum(self, a, b):
        return np.minimum(a, b)

    def argmin(self, a, axis=None):
        return np.argmin(a, axis=axis)

    def argmax(self, a, axis=None):
        return np.argmax(a, axis=axis)

    def minimum(self, a, b):
        return np.minimum(a, b)

    def maximum(self, a, b):
        return np.maximum(a, b)

    def all(self, a, axis=None, keepdims=False):
        return np.all(a, axis=axis, keepdims=keepdims)

    def sum(self, a, axis=None, keepdims=False):
        return np.sum(a, axis=axis, keepdims=keepdims)

    def prod(self, a, axis=None):
        return np.prod(a, axis=axis)

    def mean(self, a, axis=None):
        return np.mean(a, axis=axis)
    
    def median(self, a, axis=None):
        return np.median(a, axis=axis)

    def deg2rad(self, a):
        return np.deg2rad(a)

    def diff(self, a, axis=-1):
        return np.diff(a, axis=axis)

    def var(self, a, axis=None):
        return np.var(a, axis=axis)

    def cumsum(self, a, axis=None):
        return np.cumsum(a, axis=axis)

    def cross(self, a, b, axisa=-1, axisb=-1, axisc=-1, axis=None):
        return np.cross(a, b, axisa=axisa, axisb=axisb, axisc=axisc, axis=axis)

    def nansum(self, a, axis=None):
        return np.nansum(a, axis=axis)

    def histogram(self, a, bins=10, weights=None):
        return np.histogram(a, bins=bins, weights=weights)

    def array(self, x, dtype=None):
        return np.array(x, dtype=dtype)
    
    def flip(self, a, axis=None):
        return np.flip(a, axis=axis)

    def transpose(self, a, axes=None):
        return np.transpose(a, axes=axes)

    def swapaxes(self, a, axis1, axis2):
        return np.swapaxes(a, axis1, axis2)

    def moveaxis(self, a, source, destination):
        return np.moveaxis(a, source, destination)

    def stack(self, arrays, axis=0):
        return np.stack(arrays, axis=axis)
    
    def repeat(self, a, repeats, axis=None):
        return np.repeat(a,repeats, axis=axis)

    def tile(self, A, reps):
        return np.tile(A, reps)

    def concatenate(self, arrays, axis=0):
        return np.concatenate(arrays, axis=axis)

    def meshgrid(self, *xi, indexing="xy"):
        return np.meshgrid(*xi, indexing=indexing)

    def linspace(self, start, stop, num=50):
        return np.linspace(start, stop, num=num)

    def arange(self, start, stop=None, step=None):
        return np.arange(start, stop, step)

    def expand_dims(self, a, axis):
        return np.expand_dims(a, axis=axis)

    def ceil(self, x):
        return np.ceil(x)

    def floor(self, x):
        return np.floor(x)

    def modf(self, x):
        return np.modf(x)

    def round(self, x):
        return np.round(x)

    def clip(self, a, a_min, a_max):
        return np.clip(a, a_min, a_max)

    def where(self, condition, x, y):
        return np.where(condition, x, y)

    def select(self, condlist, choicelist, default=0):
        return np.select(condlist, choicelist, default)

    def logical_or(self, x1, x2):
        return np.logical_or(x1, x2)

    def logical_and(self, x1, x2):
        return np.logical_and(x1, x2)

    def squeeze(self, a, axis=None):
        return np.squeeze(a, axis=axis)

    def reshape(self, a, shape=None):
        return np.reshape(a, shape=shape)

    def ravel(self, a):
        return np.ravel(a)

    def take(self, a, indices, axis=None):
        return np.take(a, indices, axis=axis)

    def interp(self, x, xp, fp, left=None, right=None, period=None):
        return np.interp(x, xp, fp, left, right, period)

    def gather(self, a, indices):
        return a[indices]

    def shape(self, x):
        return np.shape(x)
    
    def hanning(self, x):
        return np.hanning(x)
    
    def hamming(self, x):
        return np.hamming(x)
    
    def bartlett(self, x):
        return np.bartlett(x)
    
    def convolve(self,a,v,mode='full'):
        return np.convolve(a,v,mode=mode)
    
    def pad(self,array, pad_width, mode='constant'):
        return np.pad(array, pad_width, mode=mode)

    def polyfit(self,x,y, deg):
        return np.polyfit(x,y,deg)

    class add:
        @staticmethod
        def at(a, indices, b):
            a = a.copy()
            np.add.at(a, indices, b)
            return a

    def jit(self, fun, static_argnums=None, static_argnames=None):
        return fun  # No-op

    def vmap(self, fun, in_axes, out_axes=0):
        v_axes = [(i, ax) for i, ax in enumerate(in_axes) if ax is not None]

        def vectorized_fun(*args, **kwargs):
            v_sizes = [args[i].shape[ax] for i, ax in v_axes]
            v_ax_size = v_sizes[0]
            assert all(
                [v_size == v_ax_size for v_size in v_sizes]
            ), "All vectorized axes must have the same number of elements."

            results = []
            for i in range(v_ax_size):
                new_args = [
                    args[j] if ax is None else i_at(args[j], i, ax)
                    for j, ax in enumerate(in_axes)
                ]
                results.append(fun(*new_args, **kwargs))
            results = _recombine_traceables(results)
            results = _set_out_axes(results, out_axes)
            return results

        return vectorized_fun

    def scan(self, f, init, xs):
        carry = init
        ys = []
        for x in xs:
            carry, y = f(carry, x)
            ys.append(y)
        return carry, np.stack(ys)

    def as_traceable_dataclass_obj(self, obj, data_fields, aux_fields):
        # Numpy is "traceable" by default (because there is no tracing).
        as_dataclass = dataclass(type(obj))  # Make it a dataclass, though.
        obj.__class__ = as_dataclass
        return obj


def _set_out_axes(result: np.ndarray, out_axis: int):
    result_type = type(result)
    if is_traceable_dataclass(result_type):
        data_fields = get_traceable_data_fields(result_type)
        aux_fields = get_traceable_aux_fields(result_type)
        kwargs = {field: getattr(result_type, field) for field in aux_fields}
        for field in data_fields:
            kwargs[field] = _set_out_axes(getattr(result, field), out_axis)
        return result_type(**kwargs)
    elif isinstance(result, tuple):
        return tuple([_set_out_axes(r, out_axis) for r in result])
    else:
        return np.moveaxis(result, 0, out_axis)


def _recombine_traceables(results: list):
    """Ensure that the returned value of a vmapped function is consistent with jax.

    If the results is a list of traceable dataclass objects, then they are combined into
    a single instance of that object.
    If it is a tuple, a tuple is returned, processed recursively.
    Otherwise, the results are simply returned as a numpy array."""
    if is_traceable_dataclass(type(results[0])):
        assert all([is_traceable_dataclass(type(result)) for result in results])
        result_type = type(results[0])
        data_fields = get_traceable_data_fields(result_type)
        aux_fields = get_traceable_aux_fields(result_type)
        kwargs = {field: getattr(result_type, field) for field in aux_fields}
        for field in data_fields:
            results_for_field = [getattr(result, field) for result in results]
            kwargs[field] = np.stack(results_for_field, 0)
        return result_type(**kwargs)
    elif isinstance(results[0], tuple):
        assert all([isinstance(result, tuple) for result in results])
        n_items = len(results[0])
        return tuple(
            [_recombine_traceables([r[i] for r in results]) for i in range(n_items)]
        )
    else:
        return np.array(results)



from vbeam.core import ElementGeometry, TransmittedWavefront, WaveData
from vbeam.fastmath import numpy as np
from vbeam.fastmath.traceable import traceable_dataclass
from vbeam.util.geometry.v2 import distance


@traceable_dataclass()
class STAIWavefront(TransmittedWavefront):
    def __call__(
        self,
        sender: ElementGeometry,
        point_position: np.ndarray,
        wave_data: WaveData,
    ) -> float:
        return distance(sender.position, point_position)



from typing import TypeVar


def _get_function_name(f: callable):
    return getattr(f, "__qualname__", getattr(f, "__name__"))


def renamed_kwargs(version: str, **renamed_kwargs: str):
    """Print a deprecation warning when passing an argument by name when that argument
    has been renamed in a recent version.

    >>> @renamed_kwargs("1.0.5", b="renamed_arg")
    ... def f(a, renamed_arg):
    ...   return a + renamed_arg
    >>> f(a=1, b=2)
    Deprecation warning: argument 'b' of f was renamed to 'renamed_arg' in version 1.0.5.
    3

    Using the new name or given positional arguments doesn't print the warning.
    >>> f(a=1, renamed_arg=2)
    3
    >>> f(1, 2)
    3
    """

    def decorator(f):
        def wrapped(*args, **kwargs):
            new_kwargs = {}
            for k, v in kwargs.items():
                if k in renamed_kwargs:
                    print(
                        f"Deprecation warning: argument '{k}' of "
                        f"{_get_function_name(f)} was renamed to '{renamed_kwargs[k]}' "
                        f"in version {version}."
                    )
                    k = renamed_kwargs[k]
                new_kwargs[k] = v
            return f(*args, **new_kwargs)

        return wrapped

    return decorator


if __name__ == "__main__":
    import doctest

    doctest.testmod()



"""A datastructure for data related to a transmitted wave.

The most important field of :class:`WaveData` is :attr:`source`, which represents the 
focal point of the transmitted wave. A transmitted wave is usually one of three types:

* A focused wave, where the focal point is in front of the transducer.
* A diverging wave, where the focal point is behind the transducer.
* A plane wave, which is a wave that is *"focused"* at infinity. In this case, the 
  :attr:`source` field is either ``np.inf`` (infinity) or ``None``.
"""

from typing import Callable, Optional

from vbeam.fastmath import numpy as np
from vbeam.fastmath.traceable import traceable_dataclass

identity_fn = lambda x: x  # Just return value as-is


@traceable_dataclass(("source", "azimuth", "elevation", "t0"))
class WaveData:
    """A vectorizable container of wave data: anything that is specific to a single
    transmitted wave. See this class' fields for what that could be.

    WaveData is vectorizable, meaning that it works with vmap. This way we can represent
    the wave data for all transmitted waves in a single object.

    If a WaveData object contains multiple transmitted waves then each field will have
    an additional dimension. For example, source may have the shape (64, 3) if there are
    64 transmitted waves in the dataset (each with x, y, and z source coordinates)."""

    # The location (x, y, z) of the virtual source for the transmitted wave
    source: Optional[np.ndarray] = None
    azimuth: Optional[float] = None
    elevation: Optional[float] = None
    # The time at which the transmitted wave passed through the "sender" element
    t0: Optional[float] = None

    def __getitem__(self, *args) -> "WaveData":
        """Index the wave data.

        Note that a WaveData instance may be a container of multiple vectorizable
        WaveData "objects". See WaveData's class docstring.

        >>> wave_data = WaveData(np.array([[0,0,0], [1,1,1]]), np.array([0,1]))
        >>> wave_data[1]
        WaveData(source=array([1, 1, 1]), azimuth=1, elevation=None, t0=None)
        """
        _maybe_getitem = (
            lambda attr: attr.__getitem__(*args) if attr is not None else None
        )
        return WaveData(
            _maybe_getitem(self.source),
            _maybe_getitem(self.azimuth),
            _maybe_getitem(self.elevation),
            _maybe_getitem(self.t0),
        )

    @property
    def shape(self) -> tuple:
        if self.source is not None:
            return self.source.shape[:-1]
        if self.azimuth is not None:
            return self.azimuth.shape
        if self.elevation is not None:
            return self.elevation.shape
        if self.t0 is not None:
            return self.t0.shape

    @property
    def ndim(self) -> int:
        return len(self.shape)

    def with_updates_to(
        self,
        *,
        source: Callable[[np.ndarray], np.ndarray] = identity_fn,
        azimuth: Callable[[float], float] = identity_fn,
        elevation: Callable[[float], float] = identity_fn,
        t0: Callable[[float], float] = identity_fn,
    ) -> "WaveData":
        """Return a copy with updated values for the given fields.

        If the given value for a field is a function the updated field will be that
        function applied to the current field. Example:
        >>> wave_data = WaveData(np.array([0, 0, 0]))
        >>> wave_data.with_updates_to(source=lambda x: x+1)
        WaveData(source=array([1, 1, 1]), azimuth=None, elevation=None, t0=None)

        If the given value for a field is not a function then the field will simply be
        set to that value. Example:
        >>> wave_data.with_updates_to(azimuth=1)
        WaveData(source=array([0, 0, 0]), azimuth=1, elevation=None, t0=None)
        """
        return WaveData(
            source=source(self.source) if callable(source) else source,
            azimuth=azimuth(self.azimuth) if callable(azimuth) else azimuth,
            elevation=elevation(self.elevation) if callable(elevation) else elevation,
            t0=t0(self.t0) if callable(t0) else t0,
        )



from vbeam.core import InterpolationSpace1D
from vbeam.fastmath import numpy as np
from vbeam.fastmath.traceable import traceable_dataclass


@traceable_dataclass(data_fields=("min", "d", "n", "left", "right"))
class NearestInterpolation(InterpolationSpace1D):
    """Interpolate by rounding (nearest neighbour).

    >>> interp = NearestInterpolation(0, 1, 10)
    >>> fp = np.arange(10)
    >>> x = np.array([0, 0.5, 0.51, 2.3])
    >>> interp(x, fp)
    array([0, 0, 1, 2])
    """

    min: float
    d: float
    n: int
    left: float = 0
    right: float = 0

    def __call__(self, x: np.ndarray, fp: np.ndarray) -> np.ndarray:
        index = np.round((x - self.min) / self.d)
        return np.select(
            [index < 0, index >= self.n],
            [self.left, self.right],
            fp[index.astype(int)],
        )

    @property
    def start(self) -> float:
        return self.min

    @property
    def end(self) -> float:
        return (self.d * self.n) + self.min



"""Function (:func:`.signal_for_point`) for beamforming a single point, which is 
usually repeated over all :term:`points<Point>`, :term:`receivers<Receiver>` and 
:term:`transmits<Transmit>` (or any other arbitrary dimensions).

See also:
    :mod:`vbeam.beamformers`
"""

from dataclasses import dataclass
from typing import Optional, Tuple, Union

from vbeam.core.apodization import Apodization
from vbeam.core.element_geometry import ElementGeometry
from vbeam.core.interpolation import InterpolationSpace1D
from vbeam.core.kernel_data import KernelData
from vbeam.core.speed_of_sound import SpeedOfSound
from vbeam.core.wave_data import WaveData
from vbeam.core.wavefront import (
    MultipleTransmitDistances,
    ReflectedWavefront,
    TransmittedWavefront,
)
from vbeam.fastmath import numpy as np


def signal_for_point(
    sender: ElementGeometry,
    point_position: np.ndarray,
    receiver: ElementGeometry,
    signal: np.ndarray,
    transmitted_wavefront: TransmittedWavefront,
    reflected_wavefront: ReflectedWavefront,
    speed_of_sound: Union[float, SpeedOfSound],
    wave_data: WaveData,
    interpolate: InterpolationSpace1D,
    modulation_frequency: Optional[float],
    apodization: Apodization,
) -> np.ndarray:
    """The core beamforming function. Return the delayed and interpolated signal from a
    single transmit, for a single receiver, for a single point (pixel).

    To make a full beamformer, this function should be made to run (in parallel) for a
    ll points in the image, all receiving elements, all transmitted waves, etc.

    Args:
      sender: The element or array that the transmitted wave passed through at time 0.
        It "sends" the transmitted wave.
      point_position: The point/pixel to be imaged. Is always 3D and in cartesian
        coordinates (x, y, z).
      receiver: The element that received the reflected/backscattered signal.
      signal: The recorded signal of the backscattered signal for the receiver.
      transmitted_wavefront: A wavefront model for calculating the distance that the
        transmitted wave travels before reaching the point position.
      reflected_wavefront: A wavefront model for calculating the distance that the
        reflected wave travels from a point position back to a receiver. Usually, this
        is just the euclidian distance.
      speed_of_sound: The speed-of-sound of the medium. It is used to convert from
        distance in meters returned from the wavefront models to delay in seconds. If
        it is a float, then the medium has a constant speed-of-sound (it is the same
        everywhere). It may also be an instance of :class:`SpeedOfSound` that which can
        model a heterogeneous speed-of-sound medium.
      wave_data: Data about the transmitted wave, for example the position of the
        virtual source (if applicable).
      interpolate: Return the recorded backscattered signal at the time given the
        calculated delay, by interpolation.
      modulation_frequency: Optional value used to remodulate the delayed interpolated
        signal if it was a demodulated IQ signal. If the signal is not a demodulated IQ
        signal, then modulation_frequency can be set to 0 or None.
      apodization: Apodization function that weights the returned delayed interpolated
        signal.

    Returns:
      The delayed and interpolated signal from a single transmit, for a single
      receiver, for a single point (pixel).
    """
    tx_distance = transmitted_wavefront(sender, point_position, wave_data)
    rx_distance = reflected_wavefront(point_position, receiver)

    # Potentially sample the speed-of-sound using a SpeedOfSound instance.
    if isinstance(speed_of_sound, SpeedOfSound):
        speed_of_sound = speed_of_sound.average(
            sender.position, point_position, receiver.position
        )

    delay = (tx_distance + rx_distance) / speed_of_sound - wave_data.t0
    signal = interpolate(delay, signal)

    if modulation_frequency is not None:
        signal = phase_correction(signal, delay, modulation_frequency)
    if isinstance(tx_distance, MultipleTransmitDistances):
        signal = tx_distance.aggregate_samples(signal)
    return signal * apodization(sender, point_position, receiver, wave_data)


def phase_correction(signal: float, delay: float, modulation_frequency: float):
    w0 = np.pi * 2 * modulation_frequency
    return signal * np.exp(1j * w0 * delay)


@dataclass
class SignalForPointData(KernelData):
    """Data needed for signal_for_point.

    See the docstring of signal_for_point for documentation of each field."""

    sender: ElementGeometry
    point_position: np.ndarray
    receiver: ElementGeometry
    signal: np.ndarray
    transmitted_wavefront: TransmittedWavefront
    reflected_wavefront: ReflectedWavefront
    speed_of_sound: Union[float, SpeedOfSound]
    wave_data: WaveData
    interpolate: InterpolationSpace1D
    modulation_frequency: Optional[float]
    apodization: Apodization

    def copy(self) -> "SignalForPointData":
        return SignalForPointData(**self)

    def __getitem__(self, key: str):
        return getattr(self, key)

    def keys(self) -> Tuple[str, ...]:
        return tuple(self.__dataclass_fields__.keys())



#!/usr/bin/env python
from setuptools import find_packages, setup

setup(
    name="raxterize",
    version="0.1.0",
    description="Razterization of 2D meshes using JAX",
    author="Magnus Dalen Kvalevåg",
    author_email="magnus.kvalevag@ntnu.no",
    packages=find_packages(),
)



import logging
import warnings

logger = logging.getLogger(__name__)


def get_numpy_backend():
    from .numpy_backend import NumpyBackend

    return NumpyBackend()


def get_jax_backend():
    from .jax_backend import JaxBackend

    return JaxBackend()


def get_tensorflow_backend():
    from .tensorflow_backend import TensorflowBackend

    return TensorflowBackend()


included_backends = {
    "numpy": get_numpy_backend,
    "jax": get_jax_backend,
    "tensorflow": get_tensorflow_backend,
}
backend_priority = ["jax", "numpy"]


def get_best_available_backend():
    """Attempt to import backends in order of priority and return the first one that
    succeeds."""
    for backend_name in backend_priority:
        try:
            included_backends[backend_name]()
            logger.info(f"Using backend: {backend_name}")
            if backend_name == "numpy":
                warnings.warn(
                    "The numpy backend is not optimized for performance. Consider \
installing JAX unless you are only doing simple processing."
                )
            return backend_name
        except ImportError:
            pass
    raise ImportError(
        "Unable to import any of the supported backends. Please follow the install \
instructions for vbeam for how to install the required dependencies."
    )



from vbeam.fastmath import numpy as np
from vbeam.fastmath.traceable import traceable_dataclass


@traceable_dataclass(("a", "b", "c"))
class Line:
    """A line as defined by a*x + b*z + c = 0.

    Note that only 2D lines are supported but all point-parameters and returned points
    are 3D (has x-, y-, and z-components).
    """

    a: float
    b: float
    c: float

    @staticmethod
    def passing_through(point1: np.ndarray, point2: np.ndarray) -> "Line":
        "Construct a line that passes through both point1 and point2."
        x1, y1, z1 = point1
        x2, y2, z2 = point2
        a = z1 - z2
        b = x2 - x1
        c = x1 * z2 - x2 * z1
        return Line(a, b, c)

    @staticmethod
    def from_anchor_and_angle(anchor: np.ndarray, angle: float) -> "Line":
        "Construct a line that passes through the anchor and has the given angle."
        return Line.passing_through(
            anchor, anchor + np.array([np.cos(angle), 0, np.sin(angle)])
        )

    def intersection(l1: "Line", l2: "Line") -> np.ndarray:
        "Return the point where the lines l1 and l2 intersect."
        x = (l1.b * l2.c - l2.b * l1.c) / (l1.a * l2.b - l2.a * l1.b)
        z = (l2.a * l1.c - l1.a * l2.c) / (l1.a * l2.b - l2.a * l1.b)
        return np.array([x, 0, z])

    @property
    def angle(self):
        "The angle of the line."
        return np.arctan2(self.b, self.a)

    def signed_distance(self, point: np.ndarray) -> float:
        """Return the signed distance between the point and the nearest point on the
        line.

        The sign of the distance will be positive on the left-hand side and negative on
        the right-hand side of the line.

        The y-component of the point is ignored.

        >>> vertical_line = Line.passing_through(np.array([0,0,0]), np.array([0,0,1]))
        >>> vertical_line.signed_distance(np.array([0,0,5]))
        0.0
        >>> vertical_line.signed_distance(np.array([-1,0,5]))  # The left-hand side
        1.0
        >>> vertical_line.signed_distance(np.array([1,0,5]))  # The right-hand side
        -1.0
        """
        norm = np.sqrt(self.a**2 + self.b**2)
        return (self.a * point[..., 0] + self.b * point[..., 2] + self.c) / norm



from dataclasses import field
from typing import Tuple

from vbeam.core import ElementGeometry, TransmittedWavefront, WaveData
from vbeam.fastmath import numpy as np
from vbeam.fastmath.traceable import traceable_dataclass
from vbeam.util.geometry import Line
from vbeam.wavefront import FocusedSphericalWavefront


@traceable_dataclass(("array_bounds", "base_wavefront"))
class UnifiedWavefront(TransmittedWavefront):
    """Implementation of the unified wavefront model

    https://doi.org/10.1109/tmi.2015.2456982"""

    array_bounds: Tuple[np.ndarray, np.ndarray]
    base_wavefront: TransmittedWavefront = field(
        default_factory=FocusedSphericalWavefront
    )

    def __call__(
        self,
        sender: ElementGeometry,
        point_position: np.ndarray,
        wave_data: WaveData,
    ) -> float:
        # Set up the geometry
        array_left, array_right = self.array_bounds
        line_left = Line.passing_through(array_left, wave_data.source)
        line_right = Line.passing_through(array_right, wave_data.source)
        midline = Line.from_anchor_and_angle(
            wave_data.source,
            # We subtract pi/2 because the angle is measured in reference to the x-axis
            (line_left.angle + line_right.angle) / 2 - np.pi / 2,
        )
        intersection_line = Line.from_anchor_and_angle(point_position, midline.angle)

        # The points where the scanline intersects the region boundaries
        A = line_left.intersection(intersection_line)
        B = line_right.intersection(intersection_line)

        # The weighting of the two intersection point values (used for interpolating)
        dist_A = np.sqrt(np.sum((A - point_position) ** 2))
        dist_B = np.sqrt(np.sum((B - point_position) ** 2))
        total_distance = dist_A + dist_B
        weight_A = 1 - (dist_A / total_distance)
        weight_B = 1 - (dist_B / total_distance)

        # Interpolate the distances for the intersection points of the two regions
        R1 = self.base_wavefront(sender, A, wave_data)
        R2 = self.base_wavefront(sender, B, wave_data)
        interpolated_distance = R1 * weight_A + R2 * weight_B

        # Calculate whether the point is in region I or III using XOR
        is_in_focus = (line_left.signed_distance(point_position) > 0) ^ (
            line_right.signed_distance(point_position) > 0
        )
        # Select the correct distance based on which region the point belongs to
        return np.where(
            is_in_focus,
            self.base_wavefront(sender, point_position, wave_data),
            interpolated_distance,
        )



from typing import Optional, Tuple, Union

from vbeam.apodization.window import Rectangular, Window
from vbeam.core import Apodization, ElementGeometry, WaveData
from vbeam.fastmath import numpy as np
from vbeam.fastmath.traceable import traceable_dataclass
from vbeam.util.coordinate_systems import az_el_to_cartesian
from vbeam.util.geometry.v2 import distance


@traceable_dataclass(("array_bounds", "window"))
class PlaneWaveTransmitApodization(Apodization):
    # array_bounds relative to the sender
    array_bounds: Union[
        Tuple[np.ndarray, np.ndarray],  # 1D array
        Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray],  # 2D array
    ]
    window: Optional[Window] = None

    def __post_init__(self):
        """Validate the array bounds."""
        if len(self.array_bounds) not in [2, 4]:
            raise ValueError(
                "Expected either left/right or left/right/front/back array bounds"
            )
        if not np.all(np.stack(self.array_bounds)[..., -1] == 0):
            raise NotImplementedError(
                "Currently require all array-bounds to have 0-depth"
            )
        if not np.all(np.abs(np.mean(np.array(self.array_bounds[:2]), axis=0)) < 1e-6):
            raise NotImplementedError(
                "Currently require left/right array-bounds to be centered"
            )
        if (len(self.array_bounds) == 4) and not np.all(
            np.abs(np.mean(np.array(self.array_bounds[2:]), axis=0)) < 1e-6
        ):
            raise NotImplementedError(
                "Currently require front/back array-bounds to be centered"
            )

    def __call__(
        self,
        sender: ElementGeometry,
        point_position: np.ndarray,
        receiver: ElementGeometry,
        wave_data: WaveData,
    ) -> float:
        # Normalized vector pointing in the direction of the transmitted wave
        direction = az_el_to_cartesian(
            azimuth=wave_data.azimuth, elevation=wave_data.elevation
        )

        # Project the point onto the xy-plane with origin at the beam at depth z
        point_position = point_position - sender.position
        point_x = point_position[..., 0]
        point_y = point_position[..., 1]
        point_z = point_position[..., 2]
        x_projected = point_x - direction[0] * point_z / direction[2]
        y_projected = point_y - direction[1] * point_z / direction[2]

        # Apply the window over the projected aperture and evaluate point
        window = self.window if self.window is not None else Rectangular()
        array_width = distance(self.array_bounds[1] - self.array_bounds[0])
        weight = window(np.abs(x_projected / array_width))
        if len(self.array_bounds) == 4:  # If we use a 2D probe, also apply in elevation
            array_height = distance(self.array_bounds[3] - self.array_bounds[2])
            weight *= window(np.abs(y_projected / array_height))
        return weight


@traceable_dataclass(("window", "f_number"))
class PlaneWaveReceiveApodization(Apodization):
    window: Window
    f_number: Union[float, Tuple[float, float]]

    def __call__(
        self,
        sender: ElementGeometry,
        point_position: np.ndarray,
        receiver: ElementGeometry,
        wave_data: WaveData,
    ) -> float:
        f_number = (
            self.f_number
            if isinstance(self.f_number, tuple) and len(self.f_number) == 2
            else (self.f_number, self.f_number)
        )
        dist = point_position - receiver.position
        x_dist, y_dist, z_dist = dist[0], dist[1], dist[2]
        tan_theta = x_dist / z_dist
        tan_phi = y_dist / z_dist
        ratio_theta = np.abs(f_number[0] * tan_theta)
        ratio_phi = np.abs(f_number[1] * tan_phi)
        return np.array(
            self.window(ratio_theta) * self.window(ratio_phi),
            dtype="float32",
        )



from typing import TYPE_CHECKING, Optional, Tuple, Union

from vbeam.fastmath import numpy as np
from vbeam.interpolation import FastInterpLinspace
from vbeam.util import _deprecations
from vbeam.util.arrays import grid

if TYPE_CHECKING:
    from vbeam.scan import SectorScan


def _ensure_min_and_max(min_x: float, max_x: float) -> Tuple[float, float]:
    "Swap ``min_x`` and ``max_x`` if ``min_x`` > ``max_x``."
    return tuple(
        np.where(
            min_x > max_x,
            np.array([max_x, min_x]),
            np.array([min_x, max_x]),
        )
    )


def _right_bound(
    min_azimuth: float,
    max_azimuth: float,
    min_depth: float,
    max_depth: float,
) -> float:
    """For the two arcs defined by the azimuth bounds (one for ``min_depth`` and one
    for ``max_depth``), find the right-most x coordinate of the two arcs.

    This will be the right-edge of a bounding box that encompasses the arcs. You can
    get the other edges of the bounding box by rotating the azimuth bounds by 90, 180,
    and 270 degrees.

    See ``docs/tutorials/scan/sector_scan_bounds.ipynb`` for a visualization of the
    bounding box of the arcs."""
    cos_min, cos_max = np.cos(min_azimuth), np.cos(max_azimuth)
    sin_min, sin_max = np.sin(min_azimuth), np.sin(max_azimuth)

    # Get the maximum x coordinate of the corners of both the inner and outer arc.
    max_corner_x = np.max(
        np.array(
            [
                cos_min * min_depth,  # Inner arc
                cos_max * min_depth,  # Inner arc
                cos_min * max_depth,  # Outer arc
                cos_max * max_depth,  # Outer arc
            ]
        )
    )

    # The right-most part of the arcs may either be ``max_corner_x``, or it may be on
    # the right-most *tangent* of the outer arc. We have to make some additional checks
    # to make this work. The code for this is a bit terse, so just trust the generative
    # unit tests for :attr:`SectorScan.cartesian_bounds` :)
    return np.where(
        (max_azimuth - min_azimuth) < np.pi,
        np.where(np.logical_and(sin_min < 0, sin_max > 0), max_depth, max_corner_x),
        np.where(np.logical_or(sin_min < 0, sin_max > 0), max_depth, max_corner_x),
    )


def polar_bounds_to_cartesian_bounds(
    bounds: Tuple[float, float, float, float]
) -> Tuple[float, float, float, float]:
    """Take a tuple representing the polar coordinate bounds of a grid and return the
    cartesian coordinate bounds.

    Polar coordinate bounds is of form [min_azimuth, max_azimuth, min_depth, max_depth].
    Returned cartesian coordinate bounds is of form [min_x, max_x, min_z, max_z].
    """
    min_az, max_az, min_d, max_d = bounds
    # Ensure that the min and max are actually min and max
    min_az, max_az = _ensure_min_and_max(min_az, max_az)
    min_d, max_d = _ensure_min_and_max(min_d, max_d)

    # We get the bounds by calculating the bound for each edge of the bounding box
    # individually. _right_bound gets the right-most x coordinate of the bounding
    # box and we can get the other sides by rotating the azimuth bounds by 90, 180,
    # and 270 degrees. Because in ultrasound, "straight down" is at 0 degrees, we
    # have to rotate everything by an additional 90 degrees.
    quarter_turn = np.pi / 2
    half_turn = np.pi
    # Return (left, right, top, bottom)
    return (
        -_right_bound(min_az + quarter_turn, max_az + quarter_turn, min_d, max_d),
        _right_bound(min_az - quarter_turn, max_az - quarter_turn, min_d, max_d),
        -_right_bound(min_az + half_turn, max_az + half_turn, min_d, max_d),
        _right_bound(min_az, max_az, min_d, max_d),
    )


@_deprecations.renamed_kwargs("1.0.5", imaged_points="image", sector_scan="bounds")
def scan_convert(
    image: np.ndarray,
    bounds: Union[Tuple[float, float, float, float], "SectorScan"],
    azimuth_axis: int = -2,
    depth_axis: int = -1,
    *,  # Remaining args must be passed by name (to avoid confusion)
    shape: Optional[Tuple[int, int]] = None,
    default_value: Optional[np.ndarray] = 0.0,
    edge_handling: str = "Value",
):
    from vbeam.scan import CoordinateSystem, Scan

    if isinstance(bounds, Scan):
        if not bounds.coordinate_system == CoordinateSystem.POLAR:
            raise ValueError("You may only scan convert from polar coordinates.")
        bounds = bounds.bounds
    if shape is None:
        shape = image.shape[azimuth_axis], image.shape[depth_axis]

    width, height = image.shape[azimuth_axis], image.shape[depth_axis]
    min_az, max_az, min_depth, max_depth = bounds

    # Get the points in the cartesian grid
    min_x, max_x, min_z, max_z = polar_bounds_to_cartesian_bounds(bounds)
    points = grid(
        np.linspace(min_x, max_x, shape[0]),
        np.linspace(min_z, max_z, shape[1]),
    )
    x, z = points[..., 0], points[..., 1]  # (Ignore y; scan_convert only supports 2D!)
    # and transform each point to polar coordinates.
    angles = np.arctan2(x, z)
    radii = np.sqrt(x**2 + z**2)

    # Interpolate the imaged points, sampled at the transformed points.
    return FastInterpLinspace.interp2d(
        angles,
        radii,
        FastInterpLinspace(min_az, (max_az - min_az) / (width - 1), width),
        FastInterpLinspace(min_depth, (max_depth - min_depth) / (height - 1), height),
        image,
        azimuth_axis,
        depth_axis,
        default_value=default_value,
        edge_handling=edge_handling,
    )


def parse_axes(xyz):
    "Internal utility function for parsing axes. Handles both 2D and 3D scans."
    if len(xyz) == 3:
        x, y, z = xyz
    elif len(xyz) == 2:
        x, z = xyz
        y = None
    else:
        raise ValueError(
            f"Provide either x, y, and z (3D) or only x and z (2D). Got {len(xyz)} axes"
        )
    return x, y, z



import warnings
from dataclasses import dataclass
from typing import Callable, Dict, Optional, Sequence, Union

from spekk import Spec, trees
from spekk.util.slicing import IndicesT, slice_data, slice_spec

from vbeam.apodization.plotting import plot_apodization
from vbeam.apodization.util import get_apodization_values
from vbeam.core import (
    Apodization,
    ReflectedWavefront,
    SignalForPointData,
    TransmittedWavefront,
)
from vbeam.fastmath import numpy as np
from vbeam.scan import Scan
from vbeam.scan.advanced import ExtraDimsScanMixin
from vbeam.util.transformations import *
from vbeam.wavefront.plotting import (
    plot_reflected_wavefront,
    plot_transmitted_wavefront,
)


@dataclass
class SignalForPointSetup(SignalForPointData):
    spec: Spec
    scan: Optional[Scan] = None

    @property
    def slice(self):
        class Slicer:
            def __getitem__(
                _, slices: Sequence[Union[str, IndicesT]]
            ) -> "SignalForPointSetup":
                sliced_data = slice_data(self.data, self.spec, slices)
                sliced_spec = slice_spec(self.spec, slices)
                return SignalForPointSetup(
                    **sliced_data, scan=self.scan, spec=sliced_spec
                )

        return Slicer()

    def __setattr__(self, name: str, value) -> None:
        # Handle cases where scan and point_position are set at the same time
        # Note: scan always takes precedence
        if name == "point_position" and self.scan is not None:
            raise AttributeError(
                "You may not set point_position when a scan has been defined. Try \
updating the scan instead."
            )
        if name == "scan":
            # If user is setting scan to None then that's OK. Otherwise:
            if value is not None:
                if self.point_position is not None:
                    warnings.warn("point_position will be overwritten by the scan.")

                # Try to automatically convert pyuff_ustb Scan to vbeam Scan
                try:
                    import pyuff_ustb

                    from vbeam.data_importers import parse_pyuff_scan

                    if isinstance(value, pyuff_ustb.Scan):
                        value = parse_pyuff_scan(value)
                except ModuleNotFoundError:
                    pass

                # Validate value
                if not isinstance(value, Scan):
                    raise ValueError(f"Scan must be a Scan object, got {type(value)}")
        return super().__setattr__(name, value)

    def __getattribute__(self, name: str):
        scan: Scan = object.__getattribute__(self, "scan")
        if name == "point_position" and scan is not None:
            point_position = object.__getattribute__(self, "point_position")
            if point_position is not None:
                warnings.warn(
                    "Both point_position and scan are set. Scan will be used."
                )
            value = scan.get_points()
        elif name == "spec":
            spec: Spec = object.__getattribute__(self, "spec")
            point_position = object.__getattribute__(self, "point_position")
            if scan is None and point_position is None:
                spec = spec.remove_subtree(["point_position"])
            elif isinstance(scan, ExtraDimsScanMixin):
                spec = spec.at["point_position"].set(scan.flattened_points_dimensions)
            value = spec
        else:
            value = object.__getattribute__(self, name)
        return value

    def size(
        self, dimension: Optional[Union[str, trees.Tree]] = None
    ) -> Union[int, Dict[str, int]]:
        """Return the size of the various dimensions of the data.

        ``dimension`` may be None, a string, or a tree of strings. If it is None, then
        a dict of all the dimensions and their sizes is returned. If it is a string,
        then the size of that dimension is returned. If it is a tree of strings, then
        the leaves of the tree is updated to be the size of those dimensions (assuming
        all leaves are dimensions that exists in the spec).

        See Spec.size for details."""
        data = self.data
        if dimension is None:
            # Returns a dict with all dimensions in the spec as keys and sizes as values
            return self.spec.size(data)
        # dimension may be a tree of dimensions. Update the leaves of the tree to get
        # the size of each dimension.
        return trees.update_leaves(
            dimension,
            lambda x: isinstance(x, str),
            lambda dimension: self.spec.size(data, dimension),
        )

    @property
    def data(self) -> dict:
        kernel_data_fields = SignalForPointData.__dataclass_fields__
        data = {k: getattr(self, k) for k in kernel_data_fields}
        return data

    def copy(self) -> "SignalForPointSetup":
        return SignalForPointSetup(**self.data, spec=self.spec, scan=self.scan)

    def get_apodization_values(
        self,
        dimensions: Sequence[str],
        reduce_sum_dimension: Optional[Sequence[str]] = None,
    ):
        """Return the apodization values for the dimensions. All other relevant
        dimensions are (by default) summed over.

        If you want the apodization values for each transmit and point, you'd call
        setup.get_apodization_values(["transmits", "points"]). Likewise, if you only
        want the points, you'd call setup.get_apodization_values(["points"]). All other
        dimensions are summed over."""
        return get_apodization_values(
            self.apodization,
            self.sender,
            self.point_position,
            self.receiver,
            self.wave_data,
            self.spec,
            dimensions,
            reduce_sum_dimension,
        )

    def plot_apodization(
        self,
        apodization: Optional[Apodization] = None,
        apodization_spec: Optional[Spec] = None,
        postprocess: Optional[Callable[[np.ndarray], np.ndarray]] = None,
        average: bool = True,
        jit: bool = True,
        ax=None,  # : Optional[matplotlib.pyplot.Axes]
    ):
        """Plot the apodization values using matplotlib.

        To plot just a single transmit, slice the setup object before calling, i.e.:
        `setup.slice["transmits", 20].plot_apodization()`. This also works for other
        dimensions, like "receivers", "frames", etc.

        Args:
            apodization: The apodization to plot. Defaults to the apodization set on
                this :class:`SignalForPointSetup` object.
            apodization_spec: Optional spec of the provided apodization (in case the
                apodization changes over a dimension). Defaults to None.
            postprocess: Process the apodization values further before plotting, for
                example to scan convert.
            average: Average the apodization values across dimensions instead of
                summing. Can for example average over receivers and/or transmits.
            jit: If True, the calculation of apodzation values is JIT-compiled (if the
                backend supports it).
            ax: The axis used to plot. Defaults to `matplotlib.pyplot.gca()`.
        """
        spec = self.spec.at["point_position"].set(["x", "z"])
        if apodization is None:
            apodization = self.apodization
        elif apodization_spec is not None:
            spec = spec.at["apodization"].set(apodization_spec)

        return plot_apodization(
            apodization,
            self.sender,
            self.scan.get_points(flatten=False),
            self.receiver,
            self.wave_data,
            spec,
            postprocess,
            average,
            jit,
            ax,
        )

    def plot_transmitted_wavefront(
        self,
        wavefront: Optional[TransmittedWavefront] = None,
        wavefront_spec: Optional[Spec] = None,
        postprocess: Optional[Callable[[np.ndarray], np.ndarray]] = None,
        ax=None,  # : Optional[matplotlib.pyplot.Axes]
    ):
        spec = self.spec.at["point_position"].set(["x", "z"])
        if wavefront is None:
            wavefront = self.transmitted_wavefront
        elif wavefront_spec is not None:
            spec = spec.at["transmitted_wavefront"].set(wavefront_spec)

        return plot_transmitted_wavefront(
            wavefront,
            self.sender,
            self.scan.get_points(flatten=False),
            self.wave_data,
            spec,
            postprocess,
            ax,
        )

    def plot_reflected_wavefront(
        self,
        wavefront: Optional[ReflectedWavefront] = None,
        wavefront_spec: Optional[Spec] = None,
        postprocess: Optional[Callable[[np.ndarray], np.ndarray]] = None,
        ax=None,  # : Optional[matplotlib.pyplot.Axes]
    ):
        spec = self.spec.at["point_position"].set(["x", "z"])
        if wavefront is None:
            wavefront = self.reflected_wavefront
        elif wavefront_spec is not None:
            spec = spec.at["reflected_wavefront"].set(wavefront_spec)

        return plot_reflected_wavefront(
            wavefront,
            self.scan.get_points(flatten=False),
            self.receiver,
            spec,
            postprocess,
            ax,
        )



"""A datastructure for representing a transducer element (or the full array), including 
the position, orientation, etc.
"""

from typing import Callable, Optional

from vbeam.fastmath import numpy as np
from vbeam.fastmath.traceable import traceable_dataclass

identity_fn = lambda x: x  # Just return value as-is


@traceable_dataclass(("position", "theta", "phi", "sub_elements", "parent_element"))
class ElementGeometry:
    """A vectorizable container of element geometry.

    ElementGeometry is vectorizable, meaning that it works with vmap. This way we can
    represent the element geometry for all probe elements in a single object.

    If an ElementGeometry object contains multiple elements then each field will have an
    additional dimension. For example, position may have the shape (64, 3) if there are
    64 elements (each with x, y, and z coordinates)."""

    position: np.ndarray
    theta: Optional[float] = None
    phi: Optional[float] = None
    sub_elements: Optional["ElementGeometry"] = None
    parent_element: Optional["ElementGeometry"] = None

    def __getitem__(self, *args) -> "ElementGeometry":
        """Index the element geometry.

        Note that a ElementGeometry instance may be a container of multiple vectorizable
        ElementGeometry "objects". See ElementGeometry's class docstring.

        >>> element_geometry = ElementGeometry(
        ...     np.array([[0,0,0], [1,1,1]]), np.array([0,1]), np.array([0,1])
        ... )
        >>> element_geometry[1]
        ElementGeometry(position=array([1, 1, 1]), theta=1, phi=1, sub_elements=None)
        """
        _maybe_getitem = (
            lambda attr: attr.__getitem__(*args) if attr is not None else None
        )
        return ElementGeometry(
            _maybe_getitem(self.position),
            _maybe_getitem(self.theta),
            _maybe_getitem(self.phi),
            _maybe_getitem(self.sub_elements),
            _maybe_getitem(self.parent_element),
        )

    @property
    def shape(self) -> tuple:
        return self.position.shape[:-1]

    @property
    def ndim(self) -> int:
        return len(self.shape)

    def with_updates_to(
        self,
        *,
        position: Callable[[np.ndarray], np.ndarray] = identity_fn,
        theta: Callable[[float], float] = identity_fn,
        phi: Callable[[float], float] = identity_fn,
        sub_elements: Callable[["ElementGeometry"], "ElementGeometry"] = identity_fn,
        parent_element: Callable[["ElementGeometry"], "ElementGeometry"] = identity_fn,
    ) -> "ElementGeometry":
        """Return a copy with updated values for the given fields.

        If the given value for a field is a function the updated field will be that
        function applied to the current field. Example:
        >>> element_geometry = ElementGeometry(np.array([0, 0, 0]), 1, 2)
        >>> element_geometry.with_updates_to(position=lambda x: x+1)
        ElementGeometry(position=array([1, 1, 1]), theta=1, phi=2, sub_elements=None)

        If the given value for a field is not a function then the field will simply be
        set to that value. Example:
        >>> element_geometry.with_updates_to(theta=5, phi=6)
        ElementGeometry(position=array([0, 0, 0]), theta=5, phi=6, sub_elements=None)
        """
        return ElementGeometry(
            position=position(self.position) if callable(position) else position,
            theta=theta(self.theta) if callable(theta) else theta,
            phi=phi(self.phi) if callable(phi) else phi,
            sub_elements=sub_elements(self.sub_elements)
            if callable(sub_elements)
            else sub_elements,
            parent_element=parent_element(self.parent_element)
            if callable(parent_element)
            else parent_element,
        )

    def copy(self) -> "ElementGeometry":
        return self.with_updates_to()



"""Interface for sampling the speed of sound on a line, typically the line between a 
:term:`sender` and a :term:`point`, and a :term:`point` and a :term:`receiver`. """

from abc import ABC, abstractmethod

from vbeam.fastmath import numpy as np


class SpeedOfSound(ABC):
    @abstractmethod
    def average(
        self,
        sender_position: np.ndarray,
        point_position: np.ndarray,
        receiver_position: np.ndarray,
    ) -> float:
        """Sample the speed of sound between the sender, the point position, and the
        receiver, and return the average.

        All positions are arrays of three elements: (x, y, z)."""



from typing import List, Literal, Optional, Tuple, Union

import numpy
import pyuff_ustb as pyuff
from scipy.signal import hilbert
from spekk import Spec

from vbeam.apodization import (
    Hamming,
    NoApodization,
    PlaneWaveReceiveApodization,
    PlaneWaveTransmitApodization,
    RTBApodization,
    TxRxApodization,
)
from vbeam.core import ElementGeometry, WaveData
from vbeam.data_importers.setup import SignalForPointSetup
from vbeam.fastmath import numpy as np
from vbeam.interpolation import FastInterpLinspace
from vbeam.scan import Scan, linear_scan, sector_scan
from vbeam.util.geometry.v2 import distance
from vbeam.wavefront import PlaneWavefront, ReflectedWavefront, UnifiedWavefront


def parse_pyuff_scan(scan: pyuff.Scan) -> Scan:
    "Convert a PyUFF Scan to a vbeam Scan."
    if isinstance(scan, Scan):
        return scan
    elif isinstance(scan, pyuff.LinearScan):
        return linear_scan(np.squeeze(scan.x_axis), np.squeeze(scan.z_axis))
    elif isinstance(scan, pyuff.SectorScan):
        origin = (
            scan.origin.xyz
            if isinstance(scan.origin, pyuff.Point)
            else [p.xyz for p in scan.origin]
        )
        return sector_scan(
            np.squeeze(scan.azimuth_axis),
            np.squeeze(scan.depth_axis),
            apex=np.array(origin),
        )
    else:
        raise ValueError("The scan is not an instance of pyuff.Scan")


def import_pyuff(
    channel_data: pyuff.ChannelData,
    scan: Optional[Union[pyuff.Scan, Scan]] = None,
    *,
    frames: Union[Literal["all"], int, Tuple[int], List[int], range] = "all",
) -> SignalForPointSetup:
    """Return an instance of ImportedData based on the available data in channel_data
    and optionally a pyuff scan.

    The scan may also be a vbeam.scan.Scan, in which case the resulting beamformer will
    be optimized for that scan."""
    if scan is not None:
        scan = parse_pyuff_scan(scan)

    speed_of_sound = np.array(float(channel_data.sound_speed), dtype="float32")
    t_axis_interpolate = FastInterpLinspace(
        min=float(channel_data.initial_time),
        d=1 / float(channel_data.sampling_frequency),
        n=int(channel_data.N_samples),
    )

    assert channel_data.data.ndim in (
        3,
        4,
    ), "Shape of data assumed to be either (n_samples, n_elements, n_waves) or \
(n_samples, n_elements, n_waves, n_frames). "
    # Selecting a single frame
    if isinstance(frames, int):
        if channel_data.data.ndim == 4:
            data = channel_data.data[:, :, :, frames]
        else:
            assert frames == 0, "Only frame 0 is available."
            data = channel_data.data
        receiver_signals = np.transpose(data, (2, 1, 0))
        has_multiple_frames = False
    # Selecting multiple frames
    elif isinstance(frames, (tuple, list, range)):
        if channel_data.data.ndim == 4:
            data = channel_data.data[:, :, :, frames]
        else:
            assert all(frame == 0 for frame in frames), "Only frame 0 is available."
            data = np.stack([channel_data.data for _ in frames], -1)
        receiver_signals = np.transpose(data, (3, 2, 1, 0))
        has_multiple_frames = True
    # Selecting all frames
    else:
        if channel_data.data.ndim == 4:
            receiver_signals = np.transpose(channel_data.data, (3, 2, 1, 0))
            has_multiple_frames = True
        else:
            receiver_signals = np.transpose(channel_data.data, (2, 1, 0))
            has_multiple_frames = False
    # Apply hilbert transform if modulation_frequency is 0
    modulation_frequency = np.array(channel_data.modulation_frequency)
    if numpy.abs(modulation_frequency) == 0:
        receiver_signals = np.array(hilbert(receiver_signals), dtype="complex64")

    receivers = ElementGeometry(
        np.array(channel_data.probe.xyz),
        np.array(channel_data.probe.theta),
        np.array(channel_data.probe.phi),
    )
    sender = ElementGeometry(np.array([0.0, 0.0, 0.0], dtype="float32"), 0.0, 0.0)

    sequence: List[pyuff.Wave] = channel_data.sequence
    all_wavefronts = {wave.wavefront for wave in sequence}
    assert (
        len(all_wavefronts) == 1
    ), f"There must be exactly one type of wavefront in channel_data.sequence (was \
given {all_wavefronts})."
    (wavefront,) = all_wavefronts

    array_bounds = (
        np.array([np.min(channel_data.probe.x), 0.0, 0.0]),
        np.array([np.max(channel_data.probe.x), 0.0, 0.0]),
    )

    _wave_xyz = sequence[0].source.xyz
    if wavefront == pyuff.Wavefront.plane or numpy.isinf(_wave_xyz).any():
        transmitted_wavefront = PlaneWavefront()
        apodization = TxRxApodization(
            transmit=PlaneWaveTransmitApodization(array_bounds),
            receive=PlaneWaveReceiveApodization(Hamming(), 1.7),
        )

    elif wavefront == pyuff.Wavefront.spherical:
        transmitted_wavefront = UnifiedWavefront(array_bounds)
        apodization = TxRxApodization(
            transmit=RTBApodization(array_bounds),
            receive=NoApodization(),
        )

    wave_data = WaveData(
        azimuth=np.array([wave.source.azimuth for wave in sequence]),
        elevation=np.array([wave.source.elevation for wave in sequence]),
        source=np.array([wave.source.xyz for wave in sequence]),
        t0=np.array([wave.delay for wave in sequence]),
    )
    spec = Spec(
        {
            "signal": ["transmits", "receivers", "signal_time"],
            "receiver": ["receivers"],
            "point_position": ["points"],
            "wave_data": ["transmits"],
        }
    )

    # Check if we are dealing with a STAI dataset: is each virtual source placed at
    # exactly at an element position?
    if receivers.position.shape == wave_data.source.shape and (
        numpy.allclose(receivers.position, wave_data.source)
    ):
        # We are dealing with a STAI dataset! Senders are each element in the array
        sender = receivers.copy()
        # One sending element for each transmitted wave
        spec = spec.at["sender"].set(["transmits"])
        # Redefine t0 to be when the wave passes through the sender position
        wave_data = wave_data.with_updates_to(
            t0=lambda t0: t0 - distance(sender.position) / speed_of_sound
        )

    if has_multiple_frames:
        spec = spec.add_dimension("frames", ["signal"])
    return SignalForPointSetup(
        sender=sender,
        # point_position is dynamically set from the scan in SignalForPointSetup,
        point_position=None,
        receiver=receivers,
        signal=receiver_signals,
        transmitted_wavefront=transmitted_wavefront,
        reflected_wavefront=ReflectedWavefront(),
        speed_of_sound=speed_of_sound,
        wave_data=wave_data,
        interpolate=t_axis_interpolate,
        modulation_frequency=modulation_frequency,
        apodization=apodization,
        # Additional setup
        spec=spec,
        scan=scan,
    )


def parse_beamformed_data(beamformed_data: pyuff.BeamformedData) -> np.ndarray:
    "Parse the beamformed data from a PyUFF file into an array with the correct shape."
    imaged_points = np.squeeze(beamformed_data.data)
    scan = parse_pyuff_scan(beamformed_data.scan)
    return scan.unflatten(imaged_points, points_axis=0)



"""Functions for making classes and fields traceable for the backends.

Use `traceable_dataclass` for your custom classes and `traceable` for field types that 
you know will be traceable by the active backend at runtime.

These methods sets attributes on the class or types that can be understood when calling
`np.as_traceable_dataclass_obj` on them.

A (long, but illustrative) example:
>>> from abc import ABC, abstractmethod
>>> from vbeam.fastmath import numpy as np, backend_manager
>>> from vbeam.fastmath.traceable import traceable, traceable_dataclass
>>> import tensorflow as tf
>>> class MyABC(ABC):
...     @abstractmethod
...     def do_thing(self):
...         ...
...
>>> @traceable_dataclass(data_fields=["a"])
... class Item(MyABC):
...     a: traceable(np.ndarray) = 1.0  # Note: wrapping np.ndarray with traceable is actually unneccesary because it is already wrapped in the backends.
...     def do_thing(self):
...         return 2.0
...
>>> @traceable_dataclass(data_fields=["item"])
... class Container(MyABC):
...     item: Item
...     def do_thing(self):
...         return 3.0
...
>>> backend_manager.active_backend = "tensorflow"
>>> item = Item()
>>> container = Container(item)
>>> @tf.function(jit_compile=True)
... def func(container: Container):
...     return container.item.a + container.item.do_thing() + container.do_thing()
...
>>> with tf.GradientTape() as tape:
...     tape.watch(container.item.a)
...     res = func(container)
...
>>> print(res.numpy(), tape.gradient(res, container.item.a).numpy())  # 6.0 1.0
"""
from vbeam.fastmath import numpy as np


def traceable_dataclass(data_fields=(), aux_fields=()):
    """Decorate the class as a traceable dataclass.

    A class decorated with @traceable_dataclass(...) is understood by the active backend as
    traceable. The class also gets some new attributes such that it acts like a
    dataclass.

    All attributes of the class must also be traceable.

    Args:
      data_fields: The attributes of the object that should be traced.
      aux_fields: Attributes of the object that should not be traced, but should persist
        through function calls regardless.

    Returns:
      A decorator that modifies the class' `__new__` function such that new objects are
      traceable dataclasses.

    Example:
    >>> @traceable_dataclass(data_fields=["a"])
    ... # `Item` can now be understood by backends
    ... class Item:
    ...     a: np.ndarray = 1.0
    """

    def as_traceable_dataclass_inner(cls):
        # Mark class as traceable
        setattr(cls, "__vbeam_fastmath_traceable_custom__", True)
        setattr(cls, "__vbeam_fastmath_traceable_data_fields__", data_fields)
        setattr(cls, "__vbeam_fastmath_traceable_aux_fields__", aux_fields)

        # Duck-type class as a spekk treedef
        setattr(
            cls,
            "__spekk_treedef_keys__",
            lambda self: (
                *get_traceable_data_fields(self),
                *get_traceable_aux_fields(self),
            ),
        )
        setattr(cls, "__spekk_treedef_get__", lambda self, key: getattr(self, key))
        setattr(
            cls,
            "__spekk_treedef_create__",
            lambda self, keys, values: cls(**dict(zip(keys, values))),
        )

        # Wrap __new__ with backend-specific wrapper
        original_new = cls.__new__
        cls.__new__ = lambda cls, *args, **kwargs: np.as_traceable_dataclass_obj(
            original_new(cls), data_fields, aux_fields
        )
        return cls

    return as_traceable_dataclass_inner


def traceable(t):
    """Mark the type-annotation of a field as traceable.

    The type annotations of class attributes are static once they have been defined.
    Wrapping it in `traceable` means that the active backend will know that it is
    traceable regardless of whether it was active when the class was defined or not.
    This way we can define custom classes even before we have set the active backend.

    The type annotations of class attributes are only really an issue when using
    Tensorflow (for now) as the active backend, because it performs type checks when the
    traceable object is created.

    Args:
      t: The class/type/annotation to be marked as traceable.

    Returns:
      A new class, `Traceable`, that inherits from `t`, and that the active backend
      understands is traceable."""

    class Traceable(t):
        __vbeam_fastmath_traceable__ = True

    return Traceable


def is_traceable_dataclass(cls):
    """Return True if the class has been decorated with `@traceable_dataclass(...)`."""
    return getattr(cls, "__vbeam_fastmath_traceable_custom__", False)


def is_traceable(t):
    """Return True if the type has been wrapped with `traceable`."""
    return getattr(t, "__vbeam_fastmath_traceable__", False)


def get_traceable_data_fields(cls):
    return getattr(cls, "__vbeam_fastmath_traceable_data_fields__", ())


def get_traceable_aux_fields(cls):
    return getattr(cls, "__vbeam_fastmath_traceable_aux_fields__", ())



from vbeam.fastmath import numpy as np


def as_polar(cartesian_point: np.ndarray):
    """Return a point in cartesian coordinates in its polar coordinates representation.

    NOTE: All y-values must be 0. FIXME"""
    x, y, z = cartesian_point[..., 0], cartesian_point[..., 1], cartesian_point[..., 2]
    azimuth_angles = np.arctan2(x, z)
    radii = np.sqrt(x**2 + y**2 + z**2)
    return np.stack([azimuth_angles, np.zeros(radii.shape), radii], -1)


def as_cartesian(polar_point: np.ndarray):
    "Return a point in polar coordinates in its cartesian coordinates representation."
    azimuth_angles = polar_point[..., 0]
    polar_angles = polar_point[..., 1]
    r = polar_point[..., 2]
    return np.stack(
        [
            r * np.sin(azimuth_angles) * np.cos(polar_angles),
            r * np.sin(azimuth_angles) * np.sin(polar_angles),
            r * np.cos(azimuth_angles),
        ],
        axis=-1,
    )


def applied_in_polar_coordinates(f):
    """Wrap a function that takes and returns a point in polar coordinates such that it
    takes and returns a point in cartesian coordinates instead.
    
    Usage:
    >>> import numpy as np
    >>> @applied_in_polar_coordinates
    ... def add_radius(polar_point, radius):
    ...   return polar_point + np.array([0, 0, radius])
    >>> cartesian_point = np.array([3, 0, 4])  # [x, y, z]
    >>> add_radius(cartesian_point, 1)
    array([3.6, 0. , 4.8])
    """
    def inner(p, *args, **kwargs):
        return as_cartesian(f(as_polar(p), *args, **kwargs))

    return inner


def applied_in_cartesian_coordinates(f):
    """Wrap a function that takes and returns a point in cartesian coordinates such that
    it takes and returns a point in polar coordinates instead.
    
    Usage:
    >>> import numpy as np
    >>> @applied_in_cartesian_coordinates
    ... def move_x(cartesian_point, dx):
    ...   return cartesian_point + np.array([dx, 0, 0])
    >>> polar_point = np.array([np.pi/2, 0, 1])  # [azimuth angle, polar angle, radius]
    >>> move_x(polar_point, 1)
    array([1.57079633, 0.        , 2.        ])
    """

    def inner(p, *args, **kwargs):
        return as_polar(f(as_cartesian(p), *args, **kwargs))

    return inner


def az_el_to_cartesian(
    azimuth: float, elevation: float, radius: float = 1.0
) -> np.ndarray:
    """Convert azimuth and elevation to cartesian coordinates.

    Uses ultrasound convention for azimuth and elevation.

    Args:
        azimuth: Azimuth angle in radians.
        elevation: Elevation angle in radians.
        radius: Radius of the point in cartesian coordinates.

    Returns:
        Point in cartesian coordinates.
    """
    direction = np.array(
        [
            np.sin(azimuth) * np.cos(elevation),
            np.sin(elevation),
            np.cos(azimuth) * np.cos(elevation),
        ]
    )
    return direction * radius



"""Experimental transformations. See individual classes for documentation."""


import operator
from dataclasses import dataclass
from typing import Callable, Sequence, TypeVar

import jax
import jax.numpy as jnp
from spekk import Spec
from spekk.process.transformations import Transformation


def i_at(arr: jnp.ndarray, i: int, ax: int):
    "Index arr at index i along axis ax."
    indices = (slice(None),) * ax + (i,)
    return arr[indices]


@dataclass
class AxisIndexer:
    axes: tuple  # The axis to index for each argument. None means no indexing.

    def __call__(self, args: tuple, i: int):
        return tuple(
            args[j] if ax is None else i_at(args[j], i, ax)
            for j, ax in enumerate(self.axes)
        )


T_f_result = TypeVar("T_f_result")
T_reduction_result = TypeVar("T_reduction_result")


def map_reduce(
    map_f: Callable[..., T_f_result],
    in_axes: Sequence[int],
    reduce_f: Callable[[T_reduction_result, T_f_result], T_reduction_result],
    initial_value: T_reduction_result,
    unroll=1,
) -> T_reduction_result:
    """Return a function that maps map_f over the vectorized axes of its arguments, and
    iteratively reduces the results with reduce_f.

    In most cases this will be more memory-efficient than using jax.vmap, which in turn
    may speed up the process (if the bottleneck is memory allocation).

    >>> import jax.numpy as jnp
    >>> a = jnp.array([1., 2., 3.])
    >>> f = lambda x: x+1

    >>> from operator import add, mul
    >>> f_generalized_reduced_sum = map_reduce(f, [0], add, 0)
    >>> float(f_generalized_reduced_sum(a))
    9.0

    >>> f_generalized_reduced_prod = map_reduce(f, [0], mul, 1)
    >>> float(f_generalized_reduced_prod(a))
    24.0
    """
    v_axes = [(i, ax) for i, ax in enumerate(in_axes) if ax is not None]

    def vectorized_fun(*args):
        v_sizes = [args[i].shape[ax] for i, ax in v_axes]
        v_ax_size = v_sizes[0]
        assert all(
            [v_size == v_ax_size for v_size in v_sizes]
        ), "All vectorized axes must have the same number of elements."

        indexer = AxisIndexer(in_axes)
        carry = reduce_f(initial_value, map_f(*indexer(args, 0)))
        result, _ = jax.lax.scan(
            lambda carry, x: (reduce_f(carry, map_f(*indexer(args, x))), None),
            carry,
            jnp.arange(1, v_ax_size),
            unroll=unroll,
        )
        return result

    return vectorized_fun


@dataclass
class Reduce(Transformation):
    """Reduce the values of a dimension iteratively.

    A Reduce transformation is generally a ForAll and Apply transformation combined,
    if the Apply transformation somehow aggregates the result (for example by summing
    over the vectorized axis).
    
    As a concrete example:
    ForAll("transmits") followed by Apply(np.sum, Axis("transmits") is equivalent to
    Reduce.Sum("transmits"), but using Reduce.Sum will likely allocate a lot less 
    memory, potentially at the cost of processing time.

    Args:
      dimension: The dimension to reduce over.
      reduce_f: The function to reduce with. For example operator.add for summation.
      initial_value: The initial value for the reduction. For example 0 for summation.
      unroll: The number of iterations to unroll. Unrolling the loop may make it run
        faster at the cost of compilation time.
    """

    dimension: str
    reduce_f: Callable[[T_reduction_result, T_f_result], T_reduction_result]
    initial_value: T_reduction_result
    unroll: int = 1

    def transform(self, f: Callable, input_spec: Spec, output_spec: Spec) -> Callable:
        in_axes = input_spec.index_for(self.dimension)
        if isinstance(in_axes, dict):
            in_axes = list(in_axes.values())
        return map_reduce(f, in_axes, self.reduce_f, self.initial_value, self.unroll)

    def preprocess_spec(self, spec: Spec) -> Spec:
        return spec.remove_dimension(self.dimension)

    def postprocess_spec(self, spec: Spec) -> Spec:
        return spec

    def __repr__(self) -> str:
        r = f'Reduce("{self.dimension}", {self.reduce_f}, {self.initial_value}'
        if self.unroll != 1:
            r += f", unroll={self.unroll}"
        r += ")"
        return r

    @staticmethod
    def Sum(dimension: str, unroll: int = 1) -> "Reduce":
        return Reduce(dimension, operator.add, 0, unroll)

    @staticmethod
    def Product(dimension: str, unroll: int = 1) -> "Reduce":
        return Reduce(dimension, operator.mul, 1, unroll)


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from typing import Sequence, Union

from vbeam.fastmath import numpy as np


def ensure_positive_index(n: int, index: Union[int, Sequence[int]]) -> int:
    """Take an index that may be negative, following numpy's negative index semantics,
    and ensure that it is positive. That is, if the index=-1, then it will be the last
    index of an array."""
    if isinstance(index, int):
        return n + index if index < 0 else index
    else:
        return [ensure_positive_index(n, i) for i in index]


def ensure_2d_point(point: np.ndarray, horizontal_dimension_idx: int = 0) -> np.ndarray:
    if point.shape[-1] == 2:
        return point
    elif point.shape[-1] == 3:
        return point[..., np.array([horizontal_dimension_idx, -1])]
    else:
        raise ValueError(f"Expected 2D or 3D point, got shape={point.shape}.")



from functools import wraps
from typing import Callable, Sequence, Type, TypeVar, Union

import numpy as np

T_carry = TypeVar("T_carry")
T_x = TypeVar("T_x")


def i_at(arr, i, ax):
    """Return arr with the i-th element along axis ax selected.

    >>> import numpy as np
    >>> arr = np.array([[1, 2, 3], [4, 5, 6]])
    >>> i_at(arr, 1, 0)
    array([4, 5, 6])
    >>> i_at(arr, 1, 1)
    array([2, 5])
    """
    indices = (slice(None),) * ax + (i,)
    return arr[indices]


class Backend:
    """An implementation of the operations needed by vbeam."""

    @property
    @wraps(np.ndarray)
    def ndarray(self) -> Type[np.ndarray]:
        raise NotImplementedError

    def is_ndarray(self, a):
        "Return True if a is an instance of ndarray."
        return isinstance(a, (self.ndarray, np.ndarray))

    @wraps(np.zeros)
    def zeros(self, shape, dtype=None):
        raise NotImplementedError

    @wraps(np.ones)
    def ones(self, shape, dtype=None):
        raise NotImplementedError

    @property
    @wraps(np.pi)
    def pi(self):
        raise NotImplementedError

    @wraps(np.abs)
    def abs(self, x):
        raise NotImplementedError

    @wraps(np.exp)
    def exp(self, x):
        raise NotImplementedError

    @wraps(np.log)
    def log(self, x):
        raise NotImplementedError

    @wraps(np.log10)
    def log10(self, x):
        raise NotImplementedError

    @wraps(np.conj)
    def conj(self, x):
        raise NotImplementedError
    
    @wraps(np.angle)
    def angle(self, x):
        raise NotImplementedError

    @wraps(np.sin)
    def sin(self, x):
        raise NotImplementedError

    @wraps(np.cos)
    def cos(self, x):
        raise NotImplementedError

    @wraps(np.tan)
    def tan(self, x):
        raise NotImplementedError

    @wraps(np.arcsin)
    def arcsin(self, x):
        raise NotImplementedError

    @wraps(np.arccos)
    def arccos(self, x):
        raise NotImplementedError

    @wraps(np.arctan2)
    def arctan2(self, y, x):
        raise NotImplementedError

    @wraps(np.sqrt)
    def sqrt(self, x):
        raise NotImplementedError

    @wraps(np.sign)
    def sign(self, x):
        raise NotImplementedError

    @wraps(np.nan_to_num)
    def nan_to_num(self, x, nan=0.0):
        raise NotImplementedError

    @wraps(np.min)
    def min(self, a, axis=None):
        raise NotImplementedError
    
    @wraps(np.minimum)
    def minimum(self, a, b):
        raise NotImplementedError

    @wraps(np.max)
    def max(self, a, axis=None):
        raise NotImplementedError
    
    @wraps(np.maximum)
    def maximum(self, a, b):
        raise NotImplementedError

    @wraps(np.argmin)
    def argmin(self, a, axis=None):
        raise NotImplementedError

    @wraps(np.argmax)
    def argmax(self, a, axis=None):
        raise NotImplementedError

    @wraps(np.minimum)
    def minimum(self, a, b):
        raise NotImplementedError

    @wraps(np.maximum)
    def maximum(self, a, b):
        raise NotImplementedError

    @wraps(np.sum)
    def sum(self, a, axis=None,keepdims=False):
        raise NotImplementedError

    @wraps(np.prod)
    def prod(self, a, axis=None):
        raise NotImplementedError

    @wraps(np.mean)
    def mean(self, a, axis=None):
        raise NotImplementedError

    @wraps(np.median)
    def median(self, a, axis=None):
        raise NotImplementedError

    @wraps(np.deg2rad)
    def deg2rad(self, a):
        raise NotImplementedError

    @wraps(np.diff)
    def diff(self, a, axis=None):
        raise NotImplementedError

    @wraps(np.var)
    def var(self, a, axis=None):
        raise NotImplementedError

    @wraps(np.cumsum)
    def cumsum(self, a, axis=None):
        raise NotImplementedError

    @wraps(np.cross)
    def cross(self, a, b, axisa=-1, axisb=-1, axisc=-1, axis=None):
        raise NotImplementedError

    @wraps(np.nansum)
    def nansum(self, a, axis=None):
        raise NotImplementedError

    @wraps(np.histogram)
    def histogram(self, a, bins=10, weights=None):
        raise NotImplementedError

    @wraps(np.array)
    def array(self, x, dtype=None):
        raise NotImplementedError

    @wraps(np.flip)
    def flip(self, a, axis=None):
        raise NotImplementedError

    @wraps(np.transpose)
    def transpose(self, a, axes=None):
        raise NotImplementedError

    @wraps(np.swapaxes)
    def swapaxes(self, a, axis1, axis2):
        raise NotImplementedError

    @wraps(np.moveaxis)
    def moveaxis(self, a, source, destination):
        raise NotImplementedError

    @wraps(np.stack)
    def stack(self, arrays, axis=0):
        raise NotImplementedError
    
    @wraps(np.repeat)
    def repeat(self, a, repeats, axis=None):
        raise NotImplementedError

    @wraps(np.tile)
    def tile(self, A, reps):
        raise NotImplementedError

    @wraps(np.concatenate)
    def concatenate(self, arrays, axis=0):
        raise NotImplementedError

    @wraps(np.meshgrid)
    def meshgrid(self, *xi, indexing="xy"):
        raise NotImplementedError

    @wraps(np.linspace)
    def linspace(self, start, stop, num=50):
        raise NotImplementedError

    @wraps(np.arange)
    def arange(self, start, stop=None, step=None):
        raise NotImplementedError

    @wraps(np.expand_dims)
    def expand_dims(self, a, axis):
        raise NotImplementedError

    @wraps(np.ceil)
    def ceil(self, x):
        raise NotImplementedError

    @wraps(np.floor)
    def floor(self, x):
        raise NotImplementedError

    @wraps(np.modf)
    def modf(self, x):
        raise NotImplementedError

    @wraps(np.round)
    def round(self, x):
        raise NotImplementedError

    @wraps(np.clip)
    def clip(self, a, a_min, a_max):
        raise NotImplementedError

    @wraps(np.where)
    def where(self, condition, a, b):
        raise NotImplementedError

    @wraps(np.select)
    def select(self, condlist, choicelist, default=0):
        raise NotImplementedError

    @wraps(np.logical_or)
    def logical_or(self, x1, x2):
        raise NotImplementedError

    @wraps(np.logical_or)
    def logical_and(self, x1, x2):
        raise NotImplementedError

    @wraps(np.squeeze)
    def squeeze(self, a, axis=None):
        raise NotImplementedError
    
    @wraps(np.reshape)
    def reshape(self, a, shape=None):
        raise NotImplementedError

    @wraps(np.ravel)
    def ravel(self, a):
        raise NotImplementedError

    @wraps(np.take)
    def take(self, a, indices, axis=None):
        raise NotImplementedError

    @wraps(np.interp)
    def interp(self, x, xp, fp, left=None, right=None, period=None):
        raise NotImplementedError

    @wraps(np.shape)
    def shape(self, x):
        raise NotImplementedError
    
    @wraps(np.hanning)
    def hanning(self, x):
        raise NotImplementedError
    
    @wraps(np.hamming)
    def hamming(self, x):
        raise NotImplementedError
    
    @wraps(np.bartlett)
    def bartlett(self, x):
        raise NotImplementedError
    
    @wraps(np.convolve)
    def convolve(self, a,v,mode=None):
        raise NotImplementedError
    
    @wraps(np.pad)
    def pad(self, array, pad_width, mode='constant'):
        raise NotImplementedError
    
    @wraps(np.polyfit)
    def polyfit(self, x, y, deg):
        raise NotImplementedError

    class add:
        @staticmethod
        @wraps(np.add.at)
        def at(a, indices, b):
            raise NotImplementedError

    def gather(self, a, indices):
        """Gather slices from array a, according to indices. Same as numpy slicing, but
        is sometimes required when using Tensorflow backend.

        Example:
        >>> a = np.array([0,1,2])
        >>> (a[0,-1] == np.gather(a, [0,-1])).all()
        True
        """
        raise NotImplementedError

    def jit(self, fun, static_argnums=None, static_argnames=None):
        """Just-in-time compile fun, if the backend supports it."""
        raise NotImplementedError

    def vmap(self, fun, in_axes, out_axes=0):
        """Vectorize fun over the axes specified in in_axes, possibly on the GPU, if the
        backend supports it."""
        raise NotImplementedError

    def scan(self, f, init, xs):
        """Scan over the first axis of xs, applying f to a carry and each element of xs.
        f returns the next carry state and the output for that step. the final result is
        the final carry state and the stacked outputs.

        The semantics of scan is roughly equivalent to the following Python code:
        ```
        def scan(f, init, xs):
            carry = init
            ys = []
            for x in xs:
                carry, y = f(carry, x)
                ys.append(y)
            return carry, np.stack(ys)
        ```

        scan can be used to apply a sequential reduction, potentially saving a lot of
        memory. For example, we can compute the sum of an array using scan, where only
        one result array is allocated at a time:
        >>> from vbeam.fastmath import numpy as np
        >>> def f(carry, x):
        ...     return carry + x, None
        >>> init = 0
        >>> xs = np.array([1,2,3,4,5])
        >>> carry, _ = np.scan(f, init, xs)
        >>> carry
        15

        We can also use it as a general-purpose sequential map:
        >>> def f(carry, x):
        ...     return None, x + 1  # Add 1 to each element
        >>> init = None
        >>> xs = np.array([1,2,3,4,5])
        >>> _, ys = np.scan(f, init, xs)
        >>> ys
        array([2, 3, 4, 5, 6])

        Each step will be run sequentially (not in parallel), even on the GPU. If you
        want to vectorize a function to run in parallel you should use vmap instead (if
        the backend supports it). Note that scan and vmap have very different semantics.
        """
        raise NotImplementedError

    def as_traceable_dataclass_obj(self, obj, data_fields, aux_fields):
        """Wrap/modify obj such that it is traceable by the active backend and acts like
        a dataclass.

        See vbeam.fastmath.traceable for details.

        Args:
          obj: The object to be wrapped/modified.
          data_fields: The attributes of the object that should be traced.
          aux_fields: Attributes of the object that should not be traced, but should
            persist through function calls regardless.

        Returns:
          obj, wrapped/modified.
        """
        raise NotImplementedError

    ######################
    # Dependent functions;
    # These do not need to be implemented by subclasses:

    def reduce(
        self,
        f: Callable[[T_carry, T_x], T_carry],
        xs: Union[Sequence[T_x], np.ndarray],
        init_val: T_carry,
    ):
        """Reduce over xs using f, starting with init_val.

        An example that iteratively sums each value in the list ``[1,2,3,4,5]``:

        >>> import operator
        >>> from vbeam.fastmath import numpy as np
        >>> xs = [1,2,3,4,5]
        >>> np.reduce(operator.add, xs, 0)
        15

        It is semantically equivalent to the following Python code:

        >>> def reduce(f, xs, init_val):
        ...     carry = init_val
        ...     for x in xs:
        ...         carry = f(carry, x)
        ...     return carry
        >>> reduce(operator.add, xs, 0)
        15
        """

        def scan_fn(carry, i):
            """Call reduce_fn with the given carry and the i-th element of each arg
            across the axes defined by in_axes."""
            return f(carry, i), i

        xs = self.array(xs)
        carry, _ = self.scan(scan_fn, init_val, xs)
        return carry  # The final carry is the result


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import pytest
import numpy as np

from vbeam.apodization.plane_wave import PlaneWaveTransmitApodization
from vbeam.apodization.window import Hanning
from vbeam.core import ElementGeometry, WaveData


@pytest.fixture
def array_bounds():
    """Create a simple array geometry for testing."""
    array_left = np.array([-5, 0, 0])
    array_right = np.array([5, 0, 0])
    array_bottom = np.array([0, -5, 0])
    array_top = np.array([0, 5, 0])
    return array_left, array_right, array_bottom, array_top


@pytest.fixture
def transmit_element() -> ElementGeometry:
    """Send from the center of the array."""
    return ElementGeometry(position=np.array([0, 0, 0]))


@pytest.mark.parametrize(
    "azimuth_rad,elevation_rad,test_point,expected_value",
    [
        (
            0,  # 0 degrees (straight ahead)
            0,  # 0 degrees (straight ahead)
            np.array([0, 0, 10]),  # Point directly in front
            1.0,
        ),
        (
            0,  # 0 degrees (straight ahead)
            0,  # 0 degrees (straight ahead)
            np.array([-10, 0, 10]),  # Point far left
            0.0,
        ),
        # Azimuth apodization
        (
            np.pi / 4,  # 45 degrees
            0,
            np.array([10, 0, 10]),  # Point that should be in beam at 45 deg
            1.0,
        ),
        # Elevation apodization
        (
            0,
            np.pi / 4,  # 45 degrees
            np.array([0, 10, 10]),  # Point that should be in beam at 45 deg
            1.0,
        ),
    ],
)
def test_plane_wave_transmit_apodization(
    array_bounds,
    transmit_element,
    azimuth_rad: float,
    elevation_rad: float,
    test_point,
    expected_value,
):
    # Create apodization instance
    apodization = PlaneWaveTransmitApodization(
        array_bounds=array_bounds,
        window=None,
    )

    # Create wave data with specified angle
    wave_data = WaveData(azimuth=azimuth_rad, elevation=elevation_rad)

    # Calculate apodization value
    result = apodization(
        sender=transmit_element,
        point_position=test_point,
        receiver=transmit_element,
        wave_data=wave_data,
    )

    assert result == pytest.approx(expected_value, abs=1e-6)


def test_plane_wave_transmit_apodization_with_window(array_bounds, transmit_element):
    # Create apodization instance with Hann window
    apodization = PlaneWaveTransmitApodization(
        array_bounds=array_bounds,
        window=Hanning(),
    )

    # Test point in the middle of the beam
    center_point = np.array([0, 0, 10])
    wave_data = WaveData(azimuth=0, elevation=0)

    result = apodization(
        sender=transmit_element,
        point_position=center_point,
        receiver=transmit_element,
        wave_data=wave_data,
    )

    # Center point should have maximum window value
    assert result == pytest.approx(1.0, abs=1e-6)

    # Test point near the edge of the beam
    edge_point = np.array([4.9, 0, 10])
    result = apodization(
        sender=transmit_element,
        point_position=edge_point,
        receiver=transmit_element,
        wave_data=wave_data,
    )

    # Edge point should have value between 0 and 1
    assert 0 < result < 1

    # Test point far outside the beam
    outside_point = np.array([10, 0, 10])  # Well beyond array bounds
    result = apodization(
        sender=transmit_element,
        point_position=outside_point,
        receiver=transmit_element,
        wave_data=wave_data,
    )

    # Point outside beam width should have zero apodization
    assert result == pytest.approx(0.0, abs=1e-6)


def test_invalid_direction(array_bounds, transmit_element):
    # Create apodization instance with invalid apodization_dimension
    with pytest.raises(ValueError, match="Expected either left/right.*"):
        PlaneWaveTransmitApodization(
            array_bounds=array_bounds[:1],
        )



#!/usr/bin/env python

from setuptools import setup, find_packages


setup(
    name="vbeam",
    version="1.0.8",
    description="vbeam: a fast and differentiable beamformer",
    author="Magnus Dalen Kvalevåg",
    author_email="magnus.kvalevag@ntnu.no",
    url="https://github.com/magnusdk/vbeam",
    packages=find_packages(),
    install_requires=[
        "numpy",
        "pyuff_ustb>=2.0.7",
        "scipy",
        "spekk",
    ],
)



from vbeam.core import ElementGeometry, TransmittedWavefront, WaveData
from vbeam.fastmath import numpy as np
from vbeam.fastmath.traceable import traceable_dataclass
from vbeam.wavefront.plane import PlaneWavefront


@traceable_dataclass()
class FocusedSphericalWavefront(TransmittedWavefront):
    def __call__(
        self,
        sender: ElementGeometry,
        point_position: np.ndarray,
        wave_data: WaveData,
    ) -> float:
        sender_source_dist = np.sqrt(np.sum((sender.position - wave_data.source) ** 2))
        source_point_dist = np.sqrt(np.sum((wave_data.source - point_position) ** 2))
        return sender_source_dist * np.sign(
            wave_data.source[2] - sender.position[2]
        ) - source_point_dist * np.sign(wave_data.source[2] - point_position[2])


@traceable_dataclass(("pw_margin",))
class FocusedHybridWavefront(TransmittedWavefront):
    pw_margin: float = 0.001

    def __call__(
        self,
        sender: ElementGeometry,
        point_position: np.ndarray,
        wave_data: WaveData,
    ) -> float:
        spherical_wavefront = FocusedSphericalWavefront()
        plane_wavefront = PlaneWavefront()
        wave_data.azimuth = np.arctan2(
            wave_data.source[0] - sender.position[0],
            wave_data.source[2] - sender.position[2],
        )
        wave_data.elevation = np.arctan2(
            wave_data.source[1] - sender.position[1],
            wave_data.source[2] - sender.position[2],
        )
        return np.where(
            np.abs(point_position[2] - wave_data.source[2]) > self.pw_margin,
            spherical_wavefront(sender, point_position, wave_data),
            plane_wavefront(sender, point_position, wave_data),
        )


@traceable_dataclass(("blending_power",))
class FocusedBlendedWavefront(TransmittedWavefront):
    blending_power: float = 0.5

    def __call__(
        self,
        sender: ElementGeometry,
        point_position: np.ndarray,
        wave_data: WaveData,
    ) -> float:
        spherical_wavefront = FocusedSphericalWavefront()
        plane_wavefront = PlaneWavefront()
        wave_data.azimuth = np.arctan2(
            wave_data.source[0] - sender.position[0],
            wave_data.source[2] - sender.position[2],
        )

        wave_data.elevation = np.arctan2(
            wave_data.source[1] - sender.position[1],
            wave_data.source[2] - sender.position[2],
        )
        source_point_dist = np.sqrt(np.sum((wave_data.source - point_position) ** 2))
        normalized_distance = np.clip(
            source_point_dist / np.sqrt(np.sum((wave_data.source) ** 2)), a_min=0, a_max=1
        )
        plane_distance = plane_wavefront(sender, point_position, wave_data)
        spherical_distance = spherical_wavefront(sender, point_position, wave_data)
        return (
            spherical_distance * normalized_distance** self.blending_power
            + plane_distance * (1 - normalized_distance**self.blending_power)
        )


from typing import Callable, Iterable, Optional

from vbeam.core import Apodization
from vbeam.fastmath import numpy as np
from vbeam.fastmath.traceable import traceable_dataclass


@traceable_dataclass(("apodizations",), ("combiner",))
class CombinedApodization(Apodization):
    apodizations: Iterable[Apodization]
    combiner: Optional[Callable[list[np.ndarray], np.ndarray]] = None

    def __call__(self, *args, **kwargs) -> np.ndarray:
        """Multiply the result of calling the Apodization objects.

        The product is taken along a new axis to handle broadcasting correctly.
        """
        values = [apod(*args, **kwargs) for apod in self.apodizations]
        if self.combiner is not None:
            return self.combiner(values)
        return np.prod(np.array(values), axis=0)


def combine_apodizations(
    *apodizations: Iterable[Apodization],
    combiner: Optional[Callable[[list[np.ndarray]], np.ndarray]] = None,
) -> CombinedApodization:
    """Return a new Apodization object that combines all the given apodizations.

    By default, apodizations are combined by taking the product of all their results."""
    return CombinedApodization(apodizations, combiner)



"""Module for creating reasonable default values for some arguments to 
:func:`vbeam.core.signal_for_point`. Intended for internal use only.

Each function returns the default value and its spec."""

from typing import Tuple

from spekk import Spec

from vbeam.core import ElementGeometry, WaveData
from vbeam.fastmath import numpy as np
from vbeam.util.arrays import grid


def _default_sender(spec: Spec) -> Tuple[ElementGeometry, Spec]:
    return ElementGeometry(np.array([0, 0, 0]), 0.0, 0.0), spec.at["sender"].set([])


def _default_point_position(spec: Spec) -> Tuple[np.ndarray, Spec]:
    nx, nz = 200, 200
    point_position = grid(
        np.linspace(-20e-3, 20e-3, nx),
        np.array([0.0]),
        np.linspace(0, 40e-3, nz),
        shape=[nx, nz, 3],
    )
    return point_position, spec.at["point_position"].set(["x", "z"])


def _default_receiver(spec: Spec) -> Tuple[ElementGeometry, Spec]:
    return ElementGeometry(np.array([0, 0, 0]), 0.0, 0.0), spec.at["receiver"].set([])


def _default_wave_data(spec: Spec) -> Tuple[ElementGeometry, Spec]:
    return WaveData(np.array([5e-3, 0, 20e-3]), 0.0, 0.0, 0.0), spec.at[
        "wave_data"
    ].set([])



"Point-based apodization for weighting the delayed signal."

from abc import ABC, abstractmethod

from vbeam.fastmath import numpy as np

from .element_geometry import ElementGeometry
from .wave_data import WaveData


class Apodization(ABC):
    @abstractmethod
    def __call__(
        self,
        sender: ElementGeometry,
        point_position: np.ndarray,
        receiver: ElementGeometry,
        wave_data: WaveData,
    ) -> float:
        """
        Return the weighting for the signal for the given sender, point, receiver
        position, and wave data.

        Args:
          sender: The geometry of the sender (e.g. position (x, y, z)).
          point_position: The position of the point being imaged (x, y, z).
          receiver: The geometry of the receiver (e.g. position (x, y, z)).
          wave_data: Data specific to the wave being sent (see vbeam.core.WaveData).
        """
        ...
